import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats
# choose color palette based on the quality score folder
palette_qualscore = sns.color_palette("rocket", 3)

if quality_score == '20':
    qualscore_color = palette_qualscore[0]
elif quality_score == '30':
    qualscore_color = palette_qualscore[1]
else:
    qualscore_color = palette_qualscore[2]

def single_panel_histogram(variable, caption, labels, output_tag, annotate=True):
    '''
    Plot + save a single panel histogram.

    Plots a histogram with summary statistics and a caption for a single grouping
    of samples (i.e., a single histogram panel). Use multi_panel_histogram() if
    you want multiple histograms plotted together.
    :param variable: a list, Series, or 1-d array of values to plot
    :param caption: caption to use for the figure, as a string (f-string included)
    :param labels: dictionary of labels/values to use for the x-axis, y-axis, title,
    and sample_size, where the key is the label category (x-axis, y-axis, title, sample_size)
    and the value is a string to use for the label; title will be added to a pre-defined title
    that includes the quality score and sequencing run, followed by a colon (e.g., 'Q40 Reads (sp23):');
    sample_size can be an int or string, will not be used mathematically
    :param output_tag: suffix used at the end of the plot file name that
    describes the contents of the figure; should be short, will individual words or abbreviated words
    separated by hyphens; should not include filetype
    :param annotate: Boolean; whether to annotate the plot with summary statistics, which
    includes the mean, mode, min, max, number of values (length of the input variable), and
    sample size (number of sporocarp samples described by the variable; depending on the variable,
    number of values and sample size may be equivalent)
    :return: saves plot to the summary folder of this sequencing run
    '''

    fig_width = 7  # define first, as it's used in estimate_caption_adj
    caption_text_size = 10  # also need caption text size to estimate caption adjustment

    if caption is None:
        caption_adj = 0
    else:
        def estimate_caption_adj():
            '''
            Estimates the adjustment to the size of the figure required
            to accommodate a caption of a given length.
            :return: float; size of adjustment needed in normalized
            figure coordinates
            '''
            num_char = len(caption)

            # calculate line height based on font size using golden ratio, convert to inch
            line_height = (caption_text_size*1.61803399) / 72

            # calculate characters per line (CPL) for this font size and figure width
            px_width = fig_width * 72  # convert inches to points
            char_constant = 1.76  # estimated for matplotlib default font, DejaVu Sans Serif
            cpl = px_width * (char_constant/caption_text_size)

            num_lines = np.ceil(num_char/cpl)  # round up
            print(f"The caption is {num_char} characters long and had a cpl of {cpl}; therefore"
                  f"the number of required lines is {num_lines}")
            return num_lines * line_height
        caption_adj = estimate_caption_adj()
        print(f"Adjusting the figure by {caption_adj} inches.")

    fig_height = 4.5 + caption_adj

    fig = plt.figure(figsize=(fig_width, fig_height))  # use subplots to make room for the caption below

    ax1_stop = 4  # set up ax1 to occupy 3/4 of figure space; will switch to 4/4 (all) if annotations are turned off

    # annotate first; sets subplot spacing for histogram based on T/F for annotations
    if annotate:  # annotate unless set to False
        ax2 = fig.add_subplot(1,5,5)  # create subplot to add text annotations to, will span final column
        ax2.set_axis_off()  # remove plot elements (e.g., tics, spines)
        ax2.annotate(f"{labels['x-axis']} Summary\n", xy=(0, 0.95), ha='left', va='top', fontsize=10, linespacing=1.75)
        summary_stats_str = f"mean: {np.mean(variable):.1f}\n" \
                            f"mode: {stats.mode(variable)[0][0]} (n = {stats.mode(variable)[1][0]})\n" \
                            f"min:  {np.min(variable)}\n" \
                            f"max:  {np.max(variable)}\n" \
                            f"n = {len(variable)} reads from {labels['sample_size']} samples"
        ax2.annotate(summary_stats_str, xy=(0.1, 0.89), ha='left', va='top', fontsize=9, linespacing=1.75, wrap=True)
    else:
        ax1_stop = 5

    # plot histogram
    ax1 = plt.subplot(1,5,(1,ax1_stop))  # plot overall will have 4 equal size columns, hist will span first three
    sns.histplot(variable, color=qualscore_color, kde=True, ax=ax1)  # plot hist and label
    ax1.set_ylabel(labels['y-axis'])
    ax1.set_xlabel(labels['x-axis'])
    ax1.spines['top'].set_visible(False)  # remove top plot border
    ax1.spines['right'].set_visible(False)  # remove bottom plot border

    fig.tight_layout(w_pad=0, rect=(0,(caption_adj/fig_height),1,0.95))
    fig.text(0.5, (caption_adj/fig_height), caption, wrap=True, ha='center', va='top', fontsize=caption_text_size)  # add caption
    fig.suptitle(f"Q{quality_score} Reads ({sequencing_run}): {labels['title']}", fontsize=12, fontweight='demibold')  # add title

    output_path = f"{summary_path}{run_name}{output_tag}.png"
    fig.savefig(output_path)  # save figure
    plt.close()

    if os.path.isfile(output_path):
        print(f"The histogram '{labels['title']}' has been saved: {output_path}")  # return print confirmation that figure is saved
    else:
        print(f"The histogram '{labels['title']}' could not be saved.")

    return None



from pathlib import Path
from Bio.Seq import Seq
from datetime import datetime
import re
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import pandas as pd
from climush.constants import NOPHIX_PREFIX, SEQ_FILE_GLOB, TRIMMED_PREFIX, DEREP_PREFIX, QUALFILT_PREFIX
from climush.utilities import *

def demultiplex(file_map, multiplexed_files, settings_dict, seq_platform='pacbio'):

    # REMOVE AFTER TESTING ############
    # DON'T FORGET TO UNCOMMENT THE RUN_SUBPROCESS LINE WHERE LIMA ACTUALLY RUNS!!! #################
    # file_map = fpm.copy()
    # multiplexed_files = pacbio_files
    # settings_dict = settings.copy()
    # seq_platform = 'pacbio'
    #
    # excel_df = next(file_map['config']['bc_mapping'].glob('pacbio*.xlsx'))
    # csv_df = next(file_map['config']['bc_mapping'].glob('*.csv'))
    ###################################

    # DEFINE RELEVANT FILE PATHS ####################################################################
    # get the paths needed for demux
    mapping_dir = file_map['config']['bc_mapping']  # directory containing all mapping files
    mapping_files = list(mapping_dir.glob('*'))  # get mapping files, as list (will reuse + exhaust)

    # ACCESS USER SETTINGS FROM CONFIG ##############################################################
    run_name = settings['run_details']['run_name']  # name to use for this bioinformatics run

    # CONSTRUCT REGEX CONSTANTS #####################################################################
    # column name regex; column names vary among mapping files
    FWD_COL_RE = '(fwd)|(forward)'
    REV_COL_RE = '(rev)|(reverse)'
    SAMPLE_ID_RE = '^sample'

    # CONFIRM THAT PLATFORM IS PACBIO ###############################################################
    # check that the sequence platform is set to pacbio; not suitable for others platforms right now
    if not seq_platform == 'pacbio':
        print(f'Currently, this function only accepts PacBio sequences. Not suited for demultiplexing Sanger or '
              f'Illumina sequences.\n')
        sys.exit()

    # LOCATE MAPPING FILE FOR EACH MULTIPLEXED SEQUENCE FILE #########################################
    # get all unique run queue IDs from files needing demux; will use later on
    queue_ids = set()

    # link multiplexed sequence files in needs_demux to their corresponding mapping file in config directory
    demux_mapping = {}  # key is mapping file path, value(s) is sequence file path in needs_demux dir

    for mp_file in multiplexed_files:

        # get the sequencing core's queue ID from multiplexed seq file and corresponding climush sample ID for this queue ID
        qid = re.search('^(\d{4})', mp_file.name).group(0)  # get queue ID from file name
        queue_ids.add(qid)  # add the qid to the set of all queue ids requiring demux
        cid = settings_dict['pacbio_demultiplexing']['multiplex'][qid]  # get climush - queue ID pairing from config
        qid_files = [f for f in mapping_files if
                     re.search(f'^{cid}', f.name)]  # all files in mapping files dir that match climush id

        if len(qid_files) == 1:  # should only be one mapping file; because only 1, will also be key in dict
            if qid_files[0] in demux_mapping.keys():  # if already in dict (e.g., Pool1/Pool2)
                demux_mapping[qid_files[0]].append(mp_file)  # update list of corresponding multiplexed files w/ str
            else:
                demux_mapping.update({qid_files[0]: [mp_file]})  # add queue ID and first multiplexed file as list
        elif len(qid_files) == 0:  # if there's no corresponding mapping file..
            print(f'No mapping file matching the multiplexed file {mp_file.name} was located in '
                  f'{mapping_dir.parent.name + "/" + mapping_dir.name}. Please check that a mapping file for collection '
                  f'{cid} is present in this directory, and that the CliMush collection ID ({cid}) is at the start of the '
                  f'file name (e.g., {cid + "_mapping-file.xlsx"}).\n')
        else:  # if multiple mapping files for a given queue ID
            print(f'{len(qid_files)} mapping files were located in {mapping_dir.parent.name + "/" + mapping_dir.name}. '
                  f'Please select the number corresponding to the correct file to use for the multiplexed sequencing '
                  f'file {cid}, or type \'quit\'.\n')
            right_map = prompt_print_options(qid_files)  # get right mapping file from user input
            if right_map in demux_mapping.keys():  # add to dict
                demux_mapping[right_map].append(mp_file)  # update list of corresponding multiplexed files w/ str
            else:
                demux_mapping.update({right_map: [mp_file]})  # add queue ID and first multiplexed file as list

    # CREATE A FASTA FILE OF ALL UNIQUE BARCODES USED TO MULTIPLEX ###################################
    # define subfunctions to help locate and prepare barcodes for demultiplexing
    def get_col_name(pattern, df):

        # search for columns that match the provided pattern
        matches = [c for c in df.columns if re.search(pattern, c, re.I)]

        # output depends on the number of matches returned (ideally 1)
        if len(matches) == 1:  # if one match, return that column name
            return matches[0]

        else:  # if multiple or no matches, rerun with new user-input regex

            # check number of matches, will print out different prompt, but otherwise executes same thing
            if len(matches) == 0:  # if no matches
                print(f'No column matching the pattern \'{pattern}\' was located in the input dataframe. Please try '
                      f'another regular expression that better matches a single column out of these columns in the '
                      f'dataframe: ')
            else:
                print(f'{len(matches)} column names match the pattern \'{pattern}\' in the input dataframe. Please try '
                      f'another regular expression that better matches a single column out of these columns in the '
                      f'dataframe: ')

            print_indented_list(matches)  # print out the columns of the dataframe (formatted to be indented)
            retry_pattern = input()
            return get_col_name(pattern=retry_pattern, df=df)
    def get_barcodes(map_path, primers):

        # create an empty list to add fwd/rev barcodes to; output from function then added to set from all mapping df
        fwd_barcodes = []
        rev_barcodes = []

        # read in mapping file based on file type (.xlsx, .csv, .txt)
        mapping_tabs = import_mapping_df(df_path=map_path)

        # go through each tab in the df and get the barcodes
        for tab in mapping_tabs:  # go through each tab...
            if re.search('pool', tab, re.I):  # ...only if relevant to mapping/demux

                # get the name used for the fwd/rev barcode columns in this mapping file
                fwd_col = get_col_name(FWD_COL_RE, mapping_tabs[tab])
                rev_col = get_col_name(REV_COL_RE, mapping_tabs[tab])

                # get list of values for barcodes and sample IDs
                fwd_bc = mapping_tabs[tab][fwd_col].to_list()
                rev_bc = mapping_tabs[tab][rev_col].to_list()

                # add all barcodes in the column to the total barcode list
                fwd_barcodes.extend(fwd_bc)  # add bc to the total list
                rev_barcodes.extend(rev_bc)  # use extend to prevent sublists

        # make sure that the barcodes don't contain the primer sequences; if they do, remove primer sequence from bc
        fwd_bc_noprimer = [re.sub(f'{primers["fwd_primer"]}$', '', fwd) for fwd in fwd_barcodes]
        rev_bc_noprimer = [re.sub(f'{primers["rev_primer"]}$', '', rev) for rev in rev_barcodes]

        # confirm that primers were properly removed from the barcodes
        def primers_removed_from_bc(pre_remove_list, post_remove_list, primer):

            # OBSERVED
            # get the length of each barcode after removing the primer; keep only unique length values
            len_bc_noprimer = list(set(len(f) for f in post_remove_list))

            # EXPECTED
            # get expected length of barcode w/o primer by subtracting primer length from original barcode length

            # get the length of each barcode prior to removing the primer; keep only unique length values
            len_bc_primer = list(set(len(f) for f in pre_remove_list))

            # confirm that all original barcodes (w/ primers) are the same length (should be), then calc expected
            if len(len_bc_primer) == 1:  # if all barcodes the same length prior to removing primer...
                len_bc_expected = len_bc_primer[0] - len(primer)  # ...calc single expected length
            else:  # throw error and return False if original barcodes not all same length
                print(f'ERROR. Barcodes in the mapping file should be the same length, but are not.')
                print(f'number of unique lengths of barcodes w/ primers: {len(len_bc_primer)}')
                print(f'unique lengths of barcodes w/ primers:           {len_bc_primer}\n')
                return False

            # COMPARISON
            # if all barcodes are the same length after removing primer, and they match the expected length, return True
            if (len(len_bc_noprimer) == 1) and (int(len_bc_noprimer[0]) == len_bc_expected):
                return True
            else:
                print(f'ERROR. The barcodes after removing the primer are either different lengths or do not match '
                      f'the expected length after removing the primer.')
                print(f'number of unique barcode lengths: {len(len_bc_noprimer)}')
                print(f'length of barcode w/ primer:      {len_bc_expected}')
                print(f'length(s) of barcode w/o primer:  {len_bc}\n')
                return False

        fwd_bc_ready = primers_removed_from_bc(pre_remove_list = fwd_barcodes,
                                               post_remove_list = fwd_bc_noprimer,
                                               primer = primers["fwd_primer"])
        rev_bc_ready = primers_removed_from_bc(pre_remove_list = rev_barcodes,
                                               post_remove_list = rev_bc_noprimer,
                                               primer = primers["rev_primer"])

        # if both fwd and rev barcodes are confirmed free from primers...
        if fwd_bc_ready and rev_bc_ready:
            bc_dict = {'fwd_bc': fwd_bc_noprimer,
                       'rev_bc': rev_bc_noprimer}
            return bc_dict  # export a dictionary of primer-free barcodes
        else:  # if either had an issue..
            print(f'ERROR. There was an issue removing primers from the barcode sequencing for the mapping file: '
                  f'{map_path.name}. Demultiplexing was therefore not completed. See below to determine whether the '
                  f'forward or reverse barcodes (or both) had this issue:')
            print(f'\tprimer properly removed from...')
            print(f'\t  forward barcodes: {fwd_bc_ready}')
            print(f'\t  reverse barcodes: {rev_bc_ready}')
            return sys.exit()  # exit

    # get the fwd/rev primers, since they are sometimes included in the barcode sequence and need to be removed
    primer_dict = {}
    primer_dict['fwd_primer'] = settings_dict['primers']['fwd']['sequence']['pacbio']
    primer_dict['rev_primer'] = settings_dict['primers']['rev']['sequence']['pacbio']

    # create a dict to add all barcodes to; will be used to create the unique barcode fasta
    unique_barcodes = {'fwd_bc': set(),
                       'rev_bc': set()}

    # aggregate all barcodes from the mapping files that belong to the multiplexed sequences that need demux
    for dxmap in demux_mapping:

        # get all barcodes for this seq run's mapping file, with primers removed
        seq_run_bc = get_barcodes(map_path = dxmap, primers = primer_dict)

        # add barcodes from this seq run to the overall dict of unique barcodes
        for b in ['fwd_bc', 'rev_bc']:

            # before adding, count the number of barcodes already included in the unique barcodes dict
            num_bc_overall = len(unique_barcodes[b])

            # count the number of unique barcodes that are in this current mapping file
            num_bc_mapping = len(set(seq_run_bc[b]))

            # the number of unique barcodes from this mapping file should match the number in the overall bc dict
            if num_bc_overall == 0:  # if previously empty (first addition to dict), nothing to compare
                pass  # continue to adding this mapping file's barcodes to the overall dict
            else:
                if num_bc_overall < num_bc_mapping:  # if there are additional barcodes in this mapping file...
                    extra_bc = set(seq_run_bc[b]).difference(unique_barcodes[b])  # find extra bc
                    print(f'WARNING. {len(extra_bc)} new barcode(s) added to the overall list of unique barcodes:\n'
                          f'\textra {b.split("_")[0]} barcode(s): {extra_bc}\n'
                          f'All mapping files should contain the same set of unique forward and reverse barcodes. '
                          f'Continuing with demultiplexing, but if there\'s a future error, see the mapping file'
                          f'that triggered this warning: {map.name}\n')
                elif num_bc_overall > num_bc_mapping:  # if not all possible barcodes are in this mapping file...
                    missing_bc = unique_barcodes[b].difference(set(seq_run_bc[b]))  # find missing bc
                    print(f'WARNING. {len(missing_bc)} barcode(s) were missing from this mapping file\'s barcodes:\n'
                          f'\tmissing {b.split("_")[0]} barcode(s): {missing_bc}\n'
                          f'All mapping files should contain the same set of unique forward and reverse barcodes. '
                          f'Continuing with demultiplexing, but if there\'s a future error, see the mapping file '
                          f'that triggered this warning: {map.name}\n')
                else:
                    pass  # they are equal, as expected

            # regardless of above outcome, add barcodes from this mapping file to the overall dict of unique barcodes
            unique_barcodes[b].update(seq_run_bc[b])  # use update to add list to set

    # write out fasta file of unique barcodes; must be ordered so same numbering is used every time this is run
    # create new path in pipeline output, and use this path to create fasta file name
    fasta_out_dir = mkdir_exist_ok(new_dir = 'barcodes', parent_dir = file_map['pipeline-output']['demultiplexed'])
    barcode_fasta = (fasta_out_dir / f'barcodes_{run_name}').with_suffix('.fasta')

    # create a dictionary with the barcode seq as the key and the value is the ID name (header) in the output fasta file
    unique_barcode_labels = {}
    bc_name_index = {}

    fasta_index = 0
    with open(barcode_fasta, 'wt') as fout:
        for bc_dir in unique_barcodes:  # for fwd/rev unique barcodes...

            bc_list = list(unique_barcodes[bc_dir])  # convert set to list, so it can be sorted
            bc_list.sort()  # sort barcode list, so barcode headers in fasta are numbered the same way every time

            for b,bc in enumerate(bc_list):  # for each unique barcode in the sorted list...

                bc_name = f'{bc_dir}_{b+1}'  # create barcode id w/ barcode number; +1 to account for zero-index

                # write barcode name and its sequence to the fasta file
                fout.write(f'>{bc_name}\n')  # add header
                fout.write(f'{bc}\n')  # add barcode sequence

                # add barcode name and its sequence to the unique barcode labels dictionary
                unique_barcode_labels.update({bc: bc_name})

                # add barcode name and its index to the bc name index dictionary
                bc_name_index.update({fasta_index: bc_name})
                fasta_index += 1  # have to use counter because pulling from two sources, will reset to 0 for rev bc

    # DEMULTIPLEX USING LIMA #############################################################################
    # create a directory for the output files that lima produces
    lima_output = mkdir_exist_ok(new_dir='lima_output', parent_dir=file_map['pipeline-output']['demultiplexed'])
    lima_subdirs = []  # create list of output directories to read lima output from later on

    for qid in queue_ids:

        # get the climush sequencing run ID that corresponds to this queue ID
        seq_run = settings_dict['pacbio_demultiplexing']['multiplex'][qid]

        # create an output directory for this sequencing run using the climush sequencing run ID
        seq_run_output = mkdir_exist_ok(new_dir=seq_run, parent_dir=lima_output) # create an output dir for each seq run
        lima_subdirs.append(seq_run_output)

        # located all multiplexed sequencing files that match this queue ID
        pools_to_demux = [f for f in multiplexed_files if re.search(f'^{qid}', f.name)]

        # alert if only one pool is detected
        if len(pools_to_demux) == 1:
            msg = f'WARNING. There is only one multiplexed sequencing file ({pools_to_demux[0].name}) for sequencing ' \
                  f'run {seq_run} when two are typically expected: one for each pool of sequences that were ' \
                  f'sequenced simultaneously on separate SMRT cells. Do you wish to continue with only half of the ' \
                  f'sequences from this sequencing run?'  # yes/no/quit automatically added by function
            prompt_yes_no_quit(msg)  # if yes, auto-continues; if no or quit, exits

        for pool in pools_to_demux:

            # detect pool number from the file name; remove queue ID first to avoid any confusion
            try:  # should be able to do automatically
                pool_num = re.search('\d', re.sub(f'^{qid}\.', '', pool.name)).group(0)
            except AttributeError:  # but if re.search does not return a match
                print(f'The sequencing pool number for the multiplexed file {pool.name} could not be detected in ' \
                      f'the file name. Please enter the pool number corresponding to this file as a single digit'
                      f'(i.e., no leading zeros or decimals): ')
                pool_num = input()

            # define prefix to use for lima output files; include path to the output directory for this seq run
            out_prefix = seq_run_output / f'lima-demux_{seq_run}_pool{pool_num}'

            # assemble list of commands required to run lima demultiplexing
            lima_cmd = ['lima', pool, barcode_fasta, out_prefix, '--min-score', '93', '--hifi-preset', 'ASYMMETRIC']

            run_subprocess(cli_command_list = lima_cmd, dest_dir = lima_output)

    # CREATE DICTIONARY OF BARCODE PAIRS FOR ALL SAMPLES
    sample_barcodes = {q:{'pool1': {},
                          'pool2': {}} for q in queue_ids}

    for dxmap in demux_mapping:

        # get the queue ID from the name of the multiplexed sequencing file; need it to find key in sample_barcodes
        qid = re.search('^(\d{4})', list(demux_mapping[dxmap])[0].name).group(0)

        # import the mapping file dataframe
        mapping_df = import_mapping_df(df_path=dxmap)

        # loop through each tab (= pool1/pool2) in the df to get the sample ID and its barcode combo
        for tab in mapping_df:

            # get the name of the column used for the sample ID, fwd barcode, rev barcode
            smp_col = get_col_name(pattern=SAMPLE_ID_RE, df=mapping_df[tab])
            fwd_col = get_col_name(pattern=FWD_COL_RE, df=mapping_df[tab])
            rev_col = get_col_name(pattern=REV_COL_RE, df=mapping_df[tab])

            for i in range(mapping_df[tab].shape[0]):

                # from the dataframe, get the sample ID, fwd barcode sequence, and rev barcode sequence
                smp_id = mapping_df[tab][smp_col][i]
                fwd_full_bc = mapping_df[tab][fwd_col][i]  # pull full barcode from df (may contain primer still)
                rev_full_bc = mapping_df[tab][rev_col][i]  # may also container primer

                # get the barcode matching the full barcode from the df; keep the short (no-primer) version
                fwd_bc = [f for f in unique_barcode_labels if re.search(f'^{f}', fwd_full_bc, re.I)][0]
                fwd_id = unique_barcode_labels[fwd_bc]  # use non-primer barcode to get the barcode ID
                rev_bc = [r for r in unique_barcode_labels if re.search(f'^{r}', rev_full_bc, re.I)][0]
                rev_id = unique_barcode_labels[rev_bc]

                sample_barcodes[qid][tab].update({smp_id:{}})  # add sample key
                sample_barcodes[qid][tab][smp_id].update({fwd_id:fwd_bc, rev_id:rev_bc})  # add bc IDs and seqs

    # SORT INFO FROM LIMA OUTPUT
    # set these dictionaries up the same way it was done for sample_barcodes, to distinguish different seq runs, pools, etc.
    # dictionary for read barcode combinations
    read_barcodes = {q:{'pool1': {},
                        'pool2': {}} for q in queue_ids}
    # dictionary for read sequences w/ barcodes removed
    read_seqs = {q:{'pool1': {},
                    'pool2': {}} for q in queue_ids}

    for subdir in lima_subdirs:

        # get the queue ID from the name of the multiplexed sequencing file; need it to find key in sample_barcodes
        cid = re.search('^pacbio.+', subdir.name).group(0)  # get climush ID, which will be in name of file
        mp_dict = settings_dict['pacbio_demultiplexing']['multiplex']  # make shorter for list comp below
        qid = [k for k in mp_dict if cid in mp_dict[k]][0]  # convert to qid to match dict

        for p in ['pool1', 'pool2']:

            # find the fasta file for this pool, which has the read ID and the barcode combination
            lima_fasta = [f for f in subdir.glob('*.*') if re.search(f'{subdir.name}_{p}\.fasta', f.name, re.I)][0]

            # open the fasta file and pull the indices of the forward and reverse barcodes
            with open(lima_fasta, 'r') as fin:
                fasta_txt = fin.readlines()
                for l, line in enumerate(fasta_txt):
                    if line.startswith('>'):
                        read_id = re.findall('(?<=>).+?(?=\sbc)', line, re.I)[0]
                        read_barcode_i = list(map(int, re.search('(?<=bc=)\d{1,2},\d{1,2}', line, re.I).group(0).split(',')))
                        read_bc_names = [bc_name_index[i] for i in read_barcode_i]  # convert index to name
                        read_barcodes[qid][p].update({read_id: read_bc_names})  # add read's bc names to dict
                        read_seqs[qid][p].update({read_id: fasta_txt[l+1]})  # add read sequence to dict
                    else:
                        continue

    # COMPARE LIMA BARCODE COMBINATIONS TO MAPPING FILE BARCODE COMBINATIONS

    # create an output subdirectory in demultiplexed dir for all demultiplexed reads in this bioinformatics run
    demux_dir = mkdir_exist_ok(new_dir = f'demux_{run_name}', parent_dir=file_map['pipeline-output']['demultiplexed'])

    # for each sequencing run...
    for qid in read_barcodes:

        # get the name of the sequencing run for labeling the output and read headers
        seq_run = settings['pacbio_demultiplexing']['multiplex'][qid]

        # for each pool in this sequencing run...
        for p in ['pool1', 'pool2']:

            # subset dictionary for this seq run/pool, makes more succinct code below
            sample_subdict = sample_barcodes[qid][p]

            # subset dictionary for this seq run/pool of read ID and barcode-free sequences from lima fasta output
            seq_subdict = read_seqs[qid][p]

            # for each read in this sequencing run's pool...
            for r in read_barcodes[qid][p]:

                # get a list of the barcodes found by lima in this read
                bc_list = read_barcodes[qid][p][r]

                # look for the sample ID from the mapping file that has the same pair of barcodes
                sample_ids = [s for s in sample_subdict if len(set(sample_subdict[s].keys()).difference(set(bc_list))) == 0]

                # confirm that the read's barcode combination matches only one sample ID, as it should
                if len(sample_ids) == 1:

                    # get sample ID as string, not list (now that it is confirmed there's only one)
                    sample_id = sample_ids[0]

                    # create read ID header from sequencing run, sample ID, and last part of original read header
                    full_sample_id = '_'.join([seq_run, sample_id])
                    read_id = '_'.join([full_sample_id, r.split('/')[-2]])

                    # get the barcode-free sequence for this read
                    read_seq = seq_subdict[r]

                    # write the read ID and read sequence to the sample ID demultiplexed fasta file
                    sample_fasta = (demux_dir / f'demux_{full_sample_id}').with_suffix('.fasta')
                    with open(sample_fasta, 'a') as fout:
                        fout.write(f'>{read_id}\n')
                        fout.write(f'{read_seq}\n')

                else:
                    print(f'ERROR. The barcode combination detected in read {r} from {p} of {seq_run} matched to '
                          f'{len(sample_ids)} sample IDs, when it should only match to one. This read was not '
                          f'sorted, and this error was recorded to the error table.\n')
                    barcode_error = (file_map['pipeline-output']['demultiplexed'] / 'barcode_errors').with_suffix('.tsv')

                    # if the error file doesn't yet exist, start with writing the header
                    if not barcode_error.is_file():
                        with open(barcode_error, 'a') as fout:
                            fout.write(f'sequencing_run\tsequencing_pool\tsample_ids\tread_id\tbarcode_combination\n')

                    # write out details of the read that triggered this error
                    with open(barcode_error, 'a') as fout:
                        fout.write(f'{seq_run}\t{p}\t{", ".join(sample_ids)}\t{r}\t{", ".join(bc_list)}\n')

def filter_out_phix(input_files, file_map, kmer=31, hdist=1, keep_log=True, keep_removed_seqs=True):

    prefilt_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['prefiltered'])

    run_name = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings',
                                     config_section='run_details')['run_name']

    nophix_path = mkdir_exist_ok(new_dir=f'./{NOPHIX_PREFIX}_{run_name}', parent_dir=prefilt_parent)
    phix_path = mkdir_exist_ok(new_dir=f'./{flip_prefix(NOPHIX_PREFIX)}_{run_name}', parent_dir=prefilt_parent)

    bbduk_log = make_log_file(file_name=Path('bbduk'), dest_path=prefilt_parent)

    for file in input_files:

        nophix_out = add_prefix(file_path=file, prefix=NOPHIX_PREFIX, action=None, dest_dir=nophix_path)
        phix_out = add_prefix(file_path=file, prefix=flip_prefix(NOPHIX_PREFIX), action=None, dest_dir=phix_path)

        # bbduk does not like special characters in file path
        # nophix_out_str = escape_path_special(nophix_out)
        # phix_out_str = escape_path_special(phix_out)
        # and this didn't help it work either...

        bbduk_cmd = ['bbduk.sh', f'in={file}', f'out={nophix_out}', f'out={phix_out}',
                     'ref=phix.fa', f'k={kmer}', f'hdist={hdist}', f'stats={bbduk_log}']

        try:
            run_subprocess(bbduk_cmd, dest_dir=prefilt_parent)
        except:
            continue


    if not check_for_input(nophix_path)[0]:  # this is a temporary work around to make copies that look like output while there's a curernt issue with Java
        for file in input_files:
            add_prefix(file_path=file, prefix=NOPHIX_PREFIX, action='copy', dest_dir=nophix_path)

    return nophix_path

def pair_reads(input_files):
    pairs_dict = {}
    rev_reads = []
    for file in input_files:
        if re.search('R1', file.stem, re.I):
            pairs_dict.update({file:''})
        else:
            rev_reads.append(file)

    for rev in rev_reads:
        sample_id = re.search('.+?(?=_R2)', rev.stem).group(0)
        for k in pairs_dict.keys():
            if re.search(sample_id, k.stem, re.I):
                pairs_dict[k] = rev

    return pairs_dict

def prefilter_fastx(input_files, file_map, maxn=0):
    prefilt_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['prefiltered'])

    run_name = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings',
                                     config_section='run_details')['run_name']

    noambig_path = mkdir_exist_ok(new_dir=f'./{NOAMBIG_PREFIX}_{run_name}', parent_dir=prefilt_parent)
    ambig_path = mkdir_exist_ok(new_dir=f'./{flip_prefix(NOAMBIG_PREFIX)}_{run_name}', parent_dir=prefilt_parent)

    fastq_maxns_log = make_log_file(file_name=Path('fastq_maxns'), dest_path=prefilt_parent)

    # must provide pair of forward and reverse:
    pairs_dict = pair_reads(input_files)

    for file in pairs_dict.keys():

        noambig_out_fwd = add_prefix(file_path=file, prefix=NOAMBIG_PREFIX, action=None, dest_dir=noambig_path)
        noambig_out_rev = add_prefix(file_path=pairs_dict[file], prefix=NOAMBIG_PREFIX, action=None,
                                     dest_dir=noambig_path)
        ambig_out_fwd = add_prefix(file_path=file, prefix=flip_prefix(NOAMBIG_PREFIX), action=None,
                                   dest_dir=ambig_path)
        ambig_out_rev = add_prefix(file_path=pairs_dict[file], prefix=flip_prefix(NOAMBIG_PREFIX), action=None,
                                   dest_dir=ambig_path)

        vsearch_cmd = ['vsearch', '--fastq_filter', file, '--reverse', pairs_dict[file], '--fastq_maxns', str(maxn),
                       '--fastqout', noambig_out_fwd, '--fastqout_rev', noambig_out_rev, '--fastqout_discarded',
                       ambig_out_fwd, '--fastqout_discarded_rev', ambig_out_rev]

        run_subprocess(vsearch_cmd, dest_dir=prefilt_parent)

    return noambig_path

def identify_primers(platform, config_dict):
    primer_dict = config_dict['primers']

    target_primers = {}
    primer_names = []
    for d in primer_dict.keys():
        target_primers[d] = primer_dict[d]['sequence'][platform]
        primer_names.append(primer_dict[d]['name'][platform])

    print(f'Searching for {primer_names[0]} and {primer_names[1]} in {platform.title()} reads...\n')

    return target_primers

def remove_primers(input_files, file_map, platform, paired_end=True):
    trim_primers_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['primers-trimmed'])

    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']

    # get the forward and reverse primers
    primer_dict = identify_primers(platform, config_dict=settings)
    fwd_primer = primer_dict['fwd']
    rev_primer = primer_dict['rev']

    fwd_revcomp_primer = str(Seq(fwd_primer).reverse_complement())
    rev_revcomp_primer = str(Seq(rev_primer).reverse_complement())

    trim_path = mkdir_exist_ok(new_dir=f'./{TRIMMED_PREFIX}_{run_name}', parent_dir=trim_primers_parent)
    notrim_path = mkdir_exist_ok(new_dir=f'./{flip_prefix(TRIMMED_PREFIX)}_{run_name}', parent_dir=trim_primers_parent)

    cutadapt_settings = settings['cutadapt']
    fwd_max_err = int(cutadapt_settings['max_error_rate']['fwd'])
    rev_max_err = int(cutadapt_settings['max_error_rate']['rev'])
    if fwd_max_err == rev_max_err:
        max_err = fwd_max_err
    max_untrimmed = int(cutadapt_settings['max_untrimmed'])

    if paired_end:
        paired_dict = pair_reads(input_files)

        for file in paired_dict.keys():

            fwd_read_out = add_prefix(file_path=file, prefix=TRIMMED_PREFIX, dest_dir=trim_path, action='None')
            rev_read_out = add_prefix(file_path=paired_dict[file], prefix=TRIMMED_PREFIX, dest_dir=trim_path, action='None')

            cutadapt_cmd = ['cutadapt', '--report=minimal',
                            '-a', fwd_primer, '-A', rev_primer,
                            '-n', '2', '-e', str(max_err),
                            '-o', fwd_read_out, '-p', rev_read_out,
                            file, paired_dict[file]]

            run_subprocess(cutadapt_cmd, dest_dir=trim_primers_parent)

    else:

        for file in input_files:

            trim_out = add_prefix(file_path=file, prefix=TRIMMED_PREFIX, dest_dir=trim_path, action='None')

            if cutadapt_settings['keep_untrimmed']:
                notrim_out = add_prefix(file_path=file, prefix=flip_prefix(TRIMMED_PREFIX),
                                        dest_dir=notrim_path, action='None')
                untrim_cmd = f'--untrimmed-output={notrim_out}'
            else:
                untrim_cmd = ''


            cutadapt_cmd = ['cutadapt', '--report=minimal',
                            '-a', f'^{fwd_primer}...{rev_revcomp_primer}$',
                            '-a', f'^{rev_primer}...{fwd_revcomp_primer}$',
                            '-n', '2', '-e', str(max_err), untrim_cmd, '-o', trim_out, file]

            # create or append to cutadapt stderr/stdout, all output goes to one file (not one per sample)
            run_subprocess(cutadapt_cmd, dest_dir=trim_primers_parent)

    # quantify proportion untrimmed
    with open((trim_primers_parent / 'cutadapt.out'), 'r') as fin:
        cutadapt_df = pd.read_table(fin)
        sum_in = cutadapt_df['in_reads'].sum(0)
        sum_out = cutadapt_df['out_reads'].sum(0)
        percent_lost = ((sum_in - sum_out) / (sum_in)) * 100
        if percent_lost > max_untrimmed:
            msg = f'After primer trimming, {percent_lost:.2f}% of the input reads were lost, which is ' \
                  f'above the user-defined maximum threshold of {max_untrimmed}%.\n'
            exit_process(message=msg)
        else:
            print(f'{percent_lost:.2f}% of input reads were lost to primer trimming. This is above the user-provided '
                  f'input of {max_untrimmed}%, so proceeding to next step...\n')

    return None

def merge_reads(input_files, file_map):
    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']

    if settings['quality_filtering']['illumina']['merge_reads']:

        qfilt_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['quality-filtered'])
        merge_path = mkdir_exist_ok(new_dir=f'./{MERGED_PREFIX}_{run_name}', parent_dir=qfilt_parent)
        nomerge_path = mkdir_exist_ok(new_dir=f'./{flip_prefix(MERGED_PREFIX)}_{run_name}', parent_dir=qfilt_parent)

        merge_summary = qfilt_parent / 'vsearch_merged.log'

        paired_dict = pair_reads(input_files)

        output_list = []
        for file in paired_dict.keys():

            nomerge_fwd_out = add_prefix(file_path=file, prefix=MERGED_PREFIX,
                                         dest_dir=nomerge_path, action=None)
            nomerge_rev_out = add_prefix(file_path=paired_dict[file], prefix=flip_prefix(MERGED_PREFIX),
                                         dest_dir=nomerge_path, action=None)

            merge_out_base = add_prefix(file_path=file, prefix=MERGED_PREFIX, dest_dir=merge_path, action=None)
            sample_id = re.search('.+?(?=_R1)', merge_out_base.stem).group(0)
            merge_output = (merge_out_base.parent / sample_id).with_suffix('.fastq')
            output_list.append(merge_output)

            vsearch_merge_cmd = ['vsearch', '--fastq_mergepairs', file, '--reverse', paired_dict[file],
                                 '--fastqout', merge_output,
                                 '--fastqout_notmerged_fwd', nomerge_fwd_out,
                                 '--fastqout_notmerged_rev', nomerge_rev_out,
                                 '--eetabbedout', merge_summary]

            run_subprocess(vsearch_merge_cmd, dest_dir=qfilt_parent)

        return output_list
    else:
        return input_files

def quality_filter(input_files, platform, file_map):

    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']
    qfilt_dict = settings['quality_filtering'][platform]
    min_len = qfilt_dict['min_len']
    max_len = qfilt_dict['max_len']
    premerged = settings['quality_filtering']['illumina']['merge_reads']

    if platform == 'illumina':
        max_error = qfilt_dict['max_error']
    else:
        max_error = ''

    qfilt_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['quality-filtered'])

    qfilt_path = mkdir_exist_ok(new_dir=f'./{QUALFILT_PREFIX}_{run_name}', parent_dir=qfilt_parent)
    nofilt_path = mkdir_exist_ok(new_dir=f'./{flip_prefix(QUALFILT_PREFIX)}_{run_name}', parent_dir=qfilt_parent)

    if premerged:
        for file in input_files:
            qfilt_out = add_prefix(file_path=file, prefix=QUALFILT_PREFIX,
                                           dest_dir=qfilt_path, action='None')
            nofilt_out = add_prefix(file_path=file, prefix=flip_prefix(QUALFILT_PREFIX),
                                        dest_dir=nofilt_path, action='None')

            vsearch_filt_cmd = ['vsearch', '--fastq_filter', file,
                                '--fastqout', qfilt_out,
                                '--fastqout_discarded', nofilt_out,
                                '-fastq_maxee', str(max_error), '--fastq_maxlen', str(max_len),
                                '--fastq_minlen', str(min_len), '--sizeout']

            run_subprocess(vsearch_filt_cmd, dest_dir=qfilt_parent)
    else:
        paired_dict = pair_reads(input_files)
        for file in paired_dict.keys():

            qfilt_fwd_out = add_prefix(file_path=file, prefix=QUALFILT_PREFIX,
                                       dest_dir=qfilt_path, action='None')
            nofilt_fwd_out = add_prefix(file_path=file, prefix=flip_prefix(QUALFILT_PREFIX),
                                        dest_dir=nofilt_path, action='None')

            qfilt_rev_out = add_prefix(file_path=paired_dict[file], prefix=QUALFILT_PREFIX,
                                       dest_dir=qfilt_path, action='None')
            nofilt_rev_out = add_prefix(file_path=paired_dict[file], prefix=flip_prefix(QUALFILT_PREFIX),
                                        dest_dir=nofilt_path, action='None')

            vsearch_filt_cmd = ['vsearch', '--fastq_filter', file, '--reverse', paired_dict[file],
                                '--fastqout', qfilt_fwd_out, '--fastqout_rev', qfilt_rev_out,
                                '--fastqout_discarded', nofilt_fwd_out, '--fastqout_discarded_rev', nofilt_rev_out,
                                '-fastq_maxee', str(max_error), '--fastq_maxlen', str(max_len),
                                '--fastq_minlen', str(min_len), '--sizeout']

            run_subprocess(vsearch_filt_cmd, dest_dir=qfilt_parent)

    return None

def dereplicate(input_files, derep_step, platform, file_map):

    derep_prefix = DEREP_PREFIX + '0' + str(derep_step)

    if derep_step == 1:
        out_tag = 'derep-full-length'
    else:
        out_tag = 'derep-subregions'

    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']
    min_unique = settings['quality_filtering'][platform]['min_count'][f'derep0{derep_step}']
    premerged = settings['quality_filtering']['illumina']['merge_reads']

    derep_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output'][out_tag])

    derep_path = mkdir_exist_ok(new_dir=f'./{derep_prefix}_{run_name}', parent_dir=derep_parent)

    # derep_summary = mkdir_exist_ok(new_dir=f'./{derep_prefix}_{run_name}_summaries', parent_dir=derep_parent)
    if premerged:
        pass
    else:
        input_files = [f for f in input_files if re.search('R1', f, re.I)]  # only take fwd if not merged
        # rename files to be just sample ID or keep R1 to be clear they're forward reads only?

    for file in input_files:

        # derep always outputs a fasta format, but need to provide file name as fasta format or will appear as fastq
        # even though it isn't really fastq when you open it up
        derep_output = add_prefix(file_path=file, prefix=derep_prefix, dest_dir=derep_path, action=None).with_suffix('.fasta')

        vsearch_derep_cmd = ['vsearch', '--derep_fulllength', file,
                             '--output', derep_output,'--minuniquesize', str(min_unique),
                             '--sizeout']

        run_subprocess(vsearch_derep_cmd, dest_dir=derep_parent)

    return None

def separate_subregions(input_files, file_map):

    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']

    itsx_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['separate-subregions'])

    itsx_path = mkdir_exist_ok(new_dir=f'./{ITSX_PREFIX}_{run_name}',  parent_dir=itsx_parent)

    for file in input_files:

        itsx_output_basename = add_prefix(file_path=file, prefix=ITSX_PREFIX, dest_dir=itsx_path).with_suffix('')

        itsx_command = ['ITSx', '-i', file, '-o', itsx_output_basename, '-t', 'F', '--multi_thread', 'T',
                        '--save_regions', 'all']

        run_subprocess(itsx_command, dest_dir=itsx_parent)

    return None

def concat_regions(dir_path, regions_to_concat=['ITS1', '5_8S', 'ITS2'], header_delim=';', **kwargs):
    '''
    Concatenates provided regions for each read.

    Looks in the provided directory for files containing the substring of the
    regions provided in the regions_to_concat list. For each region in the list,
    it will store read ID, region, read length, and full-length read count in a
    dictionary, where each read ID is the primary key with details stored for each
    region (read length, full-length read count). Then will combine this information
    from the regions for each read ID. The sequences from each region are concatenated
    in the order they appear in the provided regions_to_concat list. The read length of
    the final concatenated reads is the sum of the read lengths of the regions. It will
    check that the full-length read count is the same for each of the regions, which
    further confirms that the regions are from the same full-length read. Then will write
    these concatenated reads and updated headers to a new output fasta file.
    :param dir_path: directory containing the fasta files to concatenate.
    :param regions_to_concat: an ORDERED list of regions to concatenated, with the 5' region
    first and the 3' region last; defaults to ITS regions
    :param header_delim: delimiter to use for separating the components of the header in
    the concatenated fasta output; defaults to semicolon
    :return: None, but will write out a fasta file
    '''

    # REMOVE AFTER TESTING
    # dir_path = Path('./04_separated-subregions/itsx_pacbio_sporocarp-f_2023-12_Q40')
    # regions_to_concat = ['ITS1', '5_8S', 'ITS2']
    # header_delim = ';'
    ######################
    for kw in kwargs.keys():
        if re.search('suffix', kw, re.I):
            file_suffix = f'_{kwargs[kw]}'
        else:
            file_suffix = ''

    # create dictionary for acceptable input
    accepted_regions = {'(^S.+?(?=\bs)|SSU)': 'SSU',  # could be better but not going to obsess
                        'ITS.?1': 'ITS1',  # checks for possible punctuation within
                        '5.?8S': '5.8S',  # checks for possible punctuation within
                        'ITS.?2': 'ITS2',  # checks for possible punctuation within
                        '(^L.+?(?=\bs)|LSU)': 'LSU',  # similar to SSU
                        '(full)?.?ITS(?!.?\d)': 'ITS'}  # may include full, but can't have number after (w or w/o punct)

    # check input list for valid entries, and then format string to consistent format
    # returns dict now with formatted string as key, regex as value
    def create_region_dict(region_list):
        output_dict = {}
        for region_str in region_list:
            for r in accepted_regions.keys():  # compare input region to each regex in dict
                if re.search(r, region_str, re.I):  # if it does match, stop loop and return the formatted version
                    output_dict.update({accepted_regions[r]:r})
                    break
                else:
                    continue
            else:
                return print(f'FAILURE. Did not recognize the input region {region_str}. Please choose from the '
                             f'recognized regions: {list(accepted_regions.values())}\n')
        return output_dict

    region_search_dict = create_region_dict(regions_to_concat)

    # make nested dict: read ID > region > [sequence, read len, derep01_reads(?)]
    # ADD CHECK FOR MULTIPLE FILES MATCHING THE REGION IN THIS DIRECTORY
    def search_path_with_regex(dir_path, regex, return_all=False):
        found_file_paths = []
        for file in dir_path.glob('*'):
            if re.search(regex, file.name, re.I):
                found_file_paths.append(file)
            else:
                continue

        if len(found_file_paths) > 1:
            if return_all:
                return found_file_paths
            else:
                print(f'WARNING. Multiple files matching the provided regex {regex} were located in '
                      f'{dir_path.name}. Returning only the first match.\n')  # REPLACE WITH OPTIONS?
                return found_file_paths[0]
        elif len(found_file_paths) == 0:
            return print(f'FAILURE. No files matching the provided regex {regex} were located in '
                         f'{dir_path.name}. Exiting...\n')
        else:
            return found_file_paths[0]

    fastas_to_concat = [search_path_with_regex(dir_path, regex=r) for r in region_search_dict.values()]

    SAMPLE_ID_RE = '(?<=\w_)(pacbio|sanger|illumina).+?(?=\.)'
    PREFIX_RE = '^\w.+?(?=_pacbio|_sanger|_illumina)'

    sample_name = re.search(SAMPLE_ID_RE, fastas_to_concat[0].name, re.I).group(0)
    file_prefix = re.search(PREFIX_RE, fastas_to_concat[0].name, re.I).group(0)

    if list(region_search_dict.keys()) == ['ITS1', '5.8S', 'ITS2']:
        concat_name = 'full_ITS'
    else:
        concat_name = input(f'Please provide a name to assign to this combination of subregions (use '
                            f'underscores instead of spaces):\n')

    # renamed header search regex
    # WAY TO UPDATE REGEX GLOBALLY WHEN A RENAMING OCCURS?
    READ_ID_RE = '^.+?(?=;)'
    READ_REGION_RE = '(?<=\d;)\w.+?(?=;)'
    READ_LEN_RE = '(?<=len=)[0-9]{1,}(?=bp)'
    READ_COUNT_RE = '(?<=full\-len_copies=)[0-9]{1,}'

    # ADD THESE OPTIONS IN THE CONFIG FILE?
    details_to_include = ['sequence', 'region_len', 'full-len_copies']

    concat_dict = {}
    for fasta in fastas_to_concat:
        for record in SeqIO.parse(fasta, 'fasta'):
            try:
                read_id = re.search(READ_ID_RE, record.id).group(0)
                region = re.search(READ_REGION_RE, record.id, re.I).group(0)

                length = re.search(READ_LEN_RE, record.id, re.I).group(0)
                derep01_reads = re.search(READ_COUNT_RE, record.id, re.I).group(0)
                details = {k:v for k,v in zip(details_to_include, [record.seq, length, derep01_reads])}

                if read_id in concat_dict:  # if this sample is already in the dictionary...
                    concat_dict[read_id].update({region: details})  # append the new region info to it
                else:  # if it isn't yet in the dictionary...
                    concat_dict[read_id] = {region: details}  # add a key value pair for this sample
            except:
                print(f'There was an issue extracting information from the fasta header of the input file '
                      f'{fasta}. This will occur if the sequence files produced by ITSx have not yet '
                      f'been reformatted. Please run the rename_read_header function and retry.\n')
                return sys.exit()

    # concat sequences by read
    concatenated_records = []
    for read in concat_dict.keys():
        # create lists to append to each region's records to
        record_details = {k:[] for k in details_to_include}

        # pull info from each region for this read
        for region in region_search_dict.keys():
            region_dict = concat_dict[read][region]
            for det in details_to_include:
                if det in record_details:
                    record_details[det].append(region_dict[det])  # append before sum to ensure all regions there
                else:
                    record_details[det] = region_dict[det]  # append before sum to ensure all regions there

        # confirm all regions data was acquired, then combine
        def join_if_all_present(pulled_data_dict):

            joined_data = {'seq':'',
                           'id':[read, concat_name]}

            for k in pulled_data_dict.keys():
                if len(pulled_data_dict[k]) == len(regions_to_concat):
                    if any(isinstance(i, Seq) for i in pulled_data_dict[k]):
                        joined_data['seq'] = Seq('').join(pulled_data_dict[k])  # must join with empty seq
                    else:
                        if re.search('len$', k, re.I):
                            val = sum([int(x) for x in pulled_data_dict[k]])
                            joined_data['id'].append(f'{k}={val}bp')
                        elif re.search('^full\-len|count', k, re.I):
                            if len(set(pulled_data_dict[k])) == 1:
                                derep = int(pulled_data_dict[k][0])  # if all same, doesn't matter which you chose
                                joined_data['id'].append(f'{k}={derep}')  # make int so no '' surrounding num in header
                        else:
                            print(f'Could not recognized the input data type, {k}:\n'
                                  f'\t{pulled_data_dict[k]}\n')
                            return None
                else:
                    msg = f'Did not successfully pull all {k}s from the regions to concatenate ' \
                          f'({len(pulled_data_dict[k])} out of {len(regions_to_concat)} were pulled). ' \
                          f'Debug before continuing...\n'
                    print(msg)
                    return None  # REPLACE W/ sys.exit AFTER DEBUG

            return joined_data

        combined_data = join_if_all_present(record_details)
        concat_record = SeqRecord(combined_data['seq'],
                                  name = read,
                                  id = header_delim.join(combined_data['id']),
                                  description = '')
        concatenated_records.append(concat_record)

    fasta_out = dir_path / f'{file_prefix}_{sample_name}.{concat_name}{file_suffix}.fasta'
    SeqIO.write(concatenated_records, fasta_out, 'fasta')

    return None

def check_itsx_output(itsx_dir, full_len_dir, num_bp_compare, write_to_log=True, same_threshold=99):
    SAMPLE_ID_RE = '(?<=\w_)(pacbio|sanger|illumina).+?(?=\.)'
    READ_ID_RE = '[0-9]{1,}(?=\/ccs)'

    bp = int(num_bp_compare)

    # import records as dictionary, where the read ID is the key
    def name_as_key(record):
        header = record.id
        sample_read_id = header.split(';')[0]
        return sample_read_id

    # import full ITS records and LSU (for revcomp comparison)
    full_its_records = SeqIO.to_dict(SeqIO.parse(next(itsx_dir.glob('*full_ITS*')), 'fasta'), key_function=name_as_key)
    lsu_records = SeqIO.to_dict(SeqIO.parse(next(itsx_dir.glob('*LSU*')), 'fasta'), key_function=name_as_key)

    # collect the full length _ bp sequences at start, full ITS _ bp sequences at start, and LSU _ bp seqs at end
    post_itsx_diffs = {'read_id':[],
                       'full_start':[],
                       'its_start':[],
                       'lsu_end':[]}
    for record in SeqIO.parse(full_len_dir, 'fasta'):

        sample_id = re.search(SAMPLE_ID_RE, full_len_dir.name).group(0)
        read_num = re.search(READ_ID_RE, record.id).group(0)
        read_id = '_'.join([sample_id, read_num])

        renamed_record = SeqRecord(record.seq, id=read_id, description='')

        full_len_seq = renamed_record.seq
        full_its_seq = full_its_records[renamed_record.id].seq
        try:
            lsu_seq = lsu_records[renamed_record.id].seq
        except:
            lsu_seq = 'N'*bp


        post_itsx_diffs['read_id'].append(renamed_record.id)
        post_itsx_diffs['full_start'].append(str(full_len_seq[:bp]))
        post_itsx_diffs['its_start'].append(str(full_its_seq[:bp]))
        post_itsx_diffs['lsu_end'].append(str(lsu_seq[-bp:]))

    # create df from this dictionary
    diff_df = pd.DataFrame.from_dict(post_itsx_diffs)

    # compare the read slices first as full-ITS fwd, then full-LSU revcomp
    comparison_result = []  # True if same
    equal_as = []  # str describing nature of match: fwd, revcomp, no LSU, unresolved (error)
    for i in range(diff_df.shape[0]):
        full_fwd = Seq(diff_df['full_start'][i])
        its_fwd = Seq(diff_df['its_start'][i])
        lsu_end = Seq(diff_df['lsu_end'][i])
        if full_fwd == its_fwd:
            comparison_result.append(True)
            equal_as.append('forward_strand')
            continue
        else:
            if 'N' in lsu_end:
                comparison_result.append(False)
                equal_as.append('no_lsu')
            else:
                its_revcomp = lsu_end.reverse_complement()
                if full_fwd == its_revcomp:
                    comparison_result.append(True)
                    equal_as.append('reverse_comp')
                    continue
                else:
                    comparison_result.append(False)
                    equal_as.append('unresolved')

    diff_df['comparison'] = comparison_result
    diff_df['equal_as'] = equal_as

    percent_same = (sum(diff_df['comparison']) / diff_df.shape[0])*100
    revcomp_detect = ((diff_df[diff_df['equal_as'] == 'reverse_comp'].shape[0]) /sum(diff_df['comparison']))*100
    forward_detect = ((diff_df[diff_df['equal_as'] == 'forward_strand'].shape[0]) /sum(diff_df['comparison']))*100
    missing_lsu = ((diff_df[diff_df['equal_as'] == 'no_lsu'].shape[0]) /sum(diff_df['comparison']))*100

    if write_to_log:
        log_dir = itsx_dir / 'itsx_inspection_logs'
        log_dir.mkdir(exist_ok=True)
        diff_df.to_csv(log_dir / f'itsx_inspection_comparison.csv', index=False)
        with open((log_dir / 'itsx_inspection_summary.log'), 'wt') as fout:
            fout.write(f'sample: {re.search(SAMPLE_ID_RE, full_len_dir.name).group(0)}\n')
            fout.write(f'date: {datetime.today().strftime("%Y-%M-%d")}\n\n')
            fout.write(f'All full-length reads were compared to their ITS region after running ITSx. The first'
                       f' {bp} bp of the full-length reads were aligned to the first {bp} bp of their corresponding '
                       f'full ITS read that was concatenated after running ITSx. If the first {bp} bp did not '
                       f'match, then the reverse complement of the last {bp} bp of their LSU region was compared '
                       f'to the first {bp} bp of the full-length read. This is because reads are not reoriented '
                       f'until done so by ITSx. \n\n')
            fout.write(f'For {percent_same:.2f}% of these comparisons (out of {diff_df.shape[0]} total comparisons), '
                       f'the first {bp} bp of the full length read matched the first {bp} bp of either the forward '
                       f'ITS or reverse complement of the end of the LSU for that read.\n\n')
            fout.write(f'Out of the {percent_same:.2f}% that matched, {forward_detect:.2f}% were matched by comparing '
                       f'the full-length read to the start of the ITS, while {revcomp_detect:.2f}% were matched by '
                       f'comparing the full-length read to the reverse complement of the end of the LSU region.\n\n')
            if missing_lsu > 0:
                fout.write(f'{int(missing_lsu*diff_df.shape[0])} reads ({missing_lsu:.2f}% of total) did not have '
                           f'an LSU region to compare to, since ITSx could not locate the LSU for the read(s). This '
                           f'accounts for the proportion of reads that could not be matched.\n')
    else:
        if percent_same < same_threshold:
            print(f'WARNING. The percent of reads that matched to their pre-ITSx full-length read was above '
                  f'{same_threshold}%, at {percent_same}.\n')

    return None

def check_chimeras(input_files, file_map, ref=None):
    chim_ref_files = file_map['config']['reference-db']['chimera'].glob(SEQ_FILE_GLOB)

    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']

    chim_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['chimera-checked'])

    nochim_path = mkdir_exist_ok(new_dir=f'./{NOCHIM_PREFIX}_{run_name}', parent_dir=chim_parent)
    chim_path = mkdir_exist_ok(new_dir=f'./{flip_prefix(NOCHIM_PREFIX)}_{run_name}', parent_dir=chim_parent)

    uchime_log = chim_parent / 'uchime.log'

    if ref is None:
        for file in input_files:
            nochim_out = add_prefix(file_path=file, prefix=NOCHIM_PREFIX,
                                    dest_dir=nochim_path, action=None)
            chim_out = add_prefix(file_path=file, prefix=flip_prefix(NOCHIM_PREFIX),
                                  dest_dir=chim_path, action=None)

            vsearch_denovo_cmd = ['vsearch', '--uchime3_denovo', file,
                                  '--chimeras', chim_out,
                                  '--nonchimeras', nochim_out,
                                  '--uchimeout', uchime_log]

            run_subprocess(vsearch_denovo_cmd, dest_dir=chim_parent)
    else:
        refs_to_use = []
        for r in chim_ref_files:
            result = re.search(ref, r, re.I)
            if result:
                refs_to_use.append(result.group(0))
            else:
                continue

        for chim_ref in refs_to_use:
            for file in input_files:
                nochim_out = add_prefix(file_path=file, prefix=NOCHIM_PREFIX,
                                        dest_dir=chim_path, action=None)
                chim_out = add_prefix(file_path=file, prefix=flip_prefix(NOCHIM_PREFIX),
                                      dest_dir=chim_path, action=None)

                vsearch_ref_cmd = ['vsearch', '--uchime_ref', file,
                                   '--chimeras', chim_out,
                                   '--nonchimeras', nochim_out,
                                   '--uchimeout', uchime_log,
                                   '--db', chim_ref]

                run_subprocess(vsearch_ref_cmd, dest_dir=chim_parent)

    return None

def combine_all_reads(input_files, file_map):

    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']

    total_record_list = []
    for file in input_files:
        rename_read_header(file, header_delim=';')
        for record in SeqIO.parse(file, 'fasta'):
            total_record_list.append(record)

    combined_out = file_map['pipeline-output']['otus-clustered'] / f'combined_otus_{run_name}.fasta'
    SeqIO.write(total_record_list, combined_out, 'fasta')

    return combined_out

def cluster_reads(input_files, file_map):
    settings = import_config_as_dict(file_path=file_map['config']['main'], file_handle='pipeline-settings')
    run_name = settings['run_details']['run_name']

    min_threshold = settings['clustering']['min_threshold']

    clust_parent = mkdir_exist_ok(new_dir=file_map['pipeline-output']['otus-clustered'])

    # rename headers for each file to retain source of sample
    # rename_read_headers()

    # combine all reads into one fasta file
    combined_reads = combine_all_reads(input_files, file_map)

    clust_out = add_prefix(file_path=combined_reads, prefix=CLUSTER_PREFIX,
                           dest_dir=clust_parent, action=None)
    # cluster
    vsearch_clust_cmd = ['vsearch', '--cluster_fast', combined_reads, '--clusters',
                         clust_out, '--id', str(min_threshold)]

    run_subprocess(vsearch_clust_cmd, dest_dir=clust_parent)

    return None


def create_blast_db(config_dict, file_map, taxa_list=None):
    '''
    Create a reference dataset for BLAST+ search.

    :param config_dict: configuration file dictionary; required to get the list
    of reference databases to use, as defined by user in the configuration file.
    :param file_map: file mapping, from mapping.py
    :param taxa_list: option; list of taxa to include if wanting to limit search
    to a specific taxonomic group
    :return: returns path to the newly created db for blastn search; creates a custom
    database using BLAST+ makeblastdb from the command line
    '''
    tax_settings = config_dict['taxonomy']
    db_dir = file_map['config']['reference-db']

    include_genbank = tax_settings['refdb']['genbank']['include']
    include_unite = tax_settings['refdb']['unite']['include']
    include_custom = tax_settings['refdb']['custom']['include']
    include_maarjam = tax_settings['refdb']['maarjam']['include']

    included_tag = ''
    db_include_list = []
    if include_genbank:
        genbank_fasta = db_dir.glob('*genbank*')
        db_include_list.append(genbank_fasta)
        included_tag += 'G'

    if include_unite:
        unite_fasta = db_dir.glob('*unite*')
        db_include_list.append(unite_fasta)
        included_tag += 'U'

    if include_custom:
        custom_fasta = db_dir.glob('*custom*')
        db_include_list.append(custom_fasta)
        included_tag += 'C'

    if include_maarjam:
        maarjam_fasta = db_dir.glob('*maarjam*')
        db_include_list.append(maarjam_fasta)
        included_tag += 'M'

    custom_search_records = []
    for db in db_include_list:
        for record in SeqIO.parse(db, 'fasta'):
            if taxa_list is None:
                custom_search_records.append(record)
            else:
                pass  # NEED TO SEE HOW TO GET ONLY SPECIFIC TAXONOMY, NOT SURE HOW FORMATTED

    tax_out_dir = file_map['pipeline-output']['taxonomy']

    search_date = datetime.today.strftime('%Y-%M-%d')
    output_basename = f'custom-ref-{included_tag}_{search_date}'
    output_path = (tax_out_dir / output_basename).with_suffix('.fasta')

    SeqIO.write(custom_search_records, output_path, 'fasta')

    output_db = tax_out_dir / f'{output_basename}_blast'
    blast_cmd = ['makeblastdb', '-in', output_path, '-title', output_db, '-dbtype', 'nucl',
                 output_db]

    run_subprocess(blast_cmd, dest_dir = output_db)

    return output_db

def assign_taxonomy(config_dict, file_map, taxa_list=None):
    '''
    Assign taxonomy to sequences.

    :param config_dict:
    :param file_map:
    :return:
    '''
    tax_output = file_map['pipeline-output']['taxonomy']

    run_name = config_dict['run_details']['run_name']

    if config_dict['taxonomy']['algorithm'] == 'blastn':
        ref_db = create_blast_db(config_dict, file_map, taxa_list=None)

        blast_out = (tax_output / f'{run_name}').with_suffix('.txt')
        blast_cmd = ['blastn', '-query', ref_db, '-out', blast_out]

        run_subprocess(blast_cmd, dest_dir = tax_output)

    elif config_dict['taxonomy']['algorithm'] == 'rdp':
        print(f'Not sure how to set up RDP yet...')
        rdp_cmd = []

        run_subprocess(rbp_cmd, dest_dir = tax_output)

    else:
        print(f'I don\'t yet have an option for that algorithm.')

    return None


import re
import os
import datetime
from functools import wraps

# get VSEARCH version
# subprocess run
vsearch_version = os.popen('vsearch --version').readlines()[0].split(' ')[1].split('_')[0]

# get location of the mothur executable file
mothur_path = os.popen('which mothur').read().strip()
mothur_version = os.popen('mothur --version').readlines()[1].split('=')[-1]

def func_timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = datetime.datetime.now()
        func_output = func(*args, **kwargs)
        end_time = datetime.datetime.now()
        runtime = str(end_time - start_time).split('.')[0]  # round seconds down
        print(f"{func.__name__} was executed in {runtime}.\n")
        return func_output
    return wrapper()

# define single function to carry out barcode file creation
def create_barcode_fasta(barcode_csv, pool_num):
    """
    Create a fasta file of unique forward and reverse barcodes from an input mapping file that
    should contain a column for the forward barcode, reverse barcode, and sample ID. Column naming
    is flexible, as function will look for any column name with 'fwd' or 'forward', 'rev' or
    'reverse', and 'sample' and 'id', 'no', 'num', with or without space or underscore, and case
    insensitive. Barcodes may have primers attached in these tables, which will be removed, if present,
    in this function.
    :param barcode_csv: input barcode file already read in
    :param pool_num: if sequencing run was split into multiple pools, include the pool number
    :return: no direct output, but will save a [1] fasta file of the barcodes and [2] a csv of
    sample IDs and their barcode combinations for in further steps of demultiplexing.
    """
    fwd_bc_regex = re.compile('(fwd)|(forward)', re.IGNORECASE)
    rev_bc_regex = re.compile('(rev)|(reverse)', re.IGNORECASE)
    sample_regex = re.compile('(sample)(_|\s)(id|no|num)', re.IGNORECASE)
    forward_bc_col = list(filter(fwd_bc_regex.search, barcode_csv.columns))[0]
    reverse_bc_col = list(filter(rev_bc_regex.search, barcode_csv.columns))[0]
    sample_col = list(filter(sample_regex.search, barcode_csv.columns))[0]

    # get the number of unique barcodes and the number of unique samples
    num_unique_fwd = len(barcode_csv[forward_bc_col].unique())
    num_unique_rev = len(barcode_csv[reverse_bc_col].unique())
    num_unique_samples = len(barcode_csv[sample_col].unique())

    # rename column names
    barcode_csv.rename(columns={sample_col: 'sample_id',
                                forward_bc_col: 'forward',
                                reverse_bc_col: 'reverse'},
                       inplace=True)

    # remove primers from barcodes
    with open(primer_path + "primer_fwd.fasta", 'rt') as fin:
        fwd_primer = fin.readlines()[1]
    with open(primer_path + "primer_rev.fasta", 'rt') as fin:
        rev_primer = fin.readlines()[1]
    barcode_csv['forward'] = [i.split(fwd_primer)[0] for i in barcode_csv['forward']]
    barcode_csv['reverse'] = [i.split(rev_primer)[0] for i in barcode_csv['reverse']]

    # create df that contains the sample IDs with the sequences of the barcodes
    wanted_cols = ['sample_id','forward','reverse']
    sampleid_df = pd.melt(barcode_csv[wanted_cols], id_vars='sample_id', var_name='barcode_direction', value_name='sequence')
    assert (len(barcode_csv['sample_id']) == len(sampleid_df['sample_id']) / 2), 'Dataframe likely contains extra columns that' \
                                                                                 'were not be removed from the dataframe (though ' \
                                                                                 'the code should do this).'

    # create a set of barcodes (set is list of unique values)
    fwd_bc = list(set(barcode_csv['forward']))
    rev_bc = list(set(barcode_csv['reverse']))

    barcode_df = sampleid_df.drop('sample_id', axis=1).drop_duplicates(subset='sequence').sort_values(
        by='sequence').reset_index().drop('index', axis=1)

    # make sure each sequence is unique in df
    assert len(fwd_bc) + len(rev_bc) == barcode_df.shape[0]
    assert len(barcode_df['sequence'].unique()) == len(barcode_df['sequence'])

    # create headers for barcode fasta file using the indices of the *sorted* dataframe, barcode_df
    # must be sorted to produce the same file each time, which is crucial for the way Lima works
    barcode_df['barcode_header'] = None

    for i in range(barcode_df.shape[0]):
        if barcode_df['barcode_direction'][i] == 'forward':
            barcode_df['barcode_header'][i] = "bc_fwd_" + str(barcode_df.index[i])
        else:
            barcode_df['barcode_header'][i] = "bc_rev_" + str(barcode_df.index[i])

    # write out barcodes and headers as fasta file
    fasta_output_name = output_path + run_name + "pool" + str(pool_num) + "_barcodes.fasta"
    with open(fasta_output_name, 'wt') as fout:
        for i in range(barcode_df.shape[0]):
            fout.write(">" + barcode_df['barcode_header'][i] + '\n')
            fout.write(barcode_df['sequence'][i] + '\n')

    if os.path.isfile(fasta_output_name) == True:
        print(f"\nBarcode fasta file for pool {pool_num}, {os.path.basename(fasta_output_name)}, has been created.\n")

    # merge sample id dataframe and barcode id dataframe by sequence
    merged_df = pd.merge(barcode_df, sampleid_df)
    wider_df = pd.pivot(data=merged_df, index='sample_id', columns='barcode_direction', values='barcode_header').reset_index()
    wider_df['fwd_index'] = [wider_df['forward'][i].split('_')[2] for i in range(wider_df.shape[0])]
    wider_df['rev_index'] = [wider_df['reverse'][i].split('_')[2] for i in range(wider_df.shape[0])]

    csv_output_name = output_path + run_name + "pool" + str(pool_num) + "_merged.csv"
    wider_df.to_csv(csv_output_name, index=False)

    assert len(wider_df['forward'].unique()) == num_unique_fwd
    assert len(wider_df['reverse'].unique()) == num_unique_rev
    assert len(wider_df.index.unique()) == num_unique_samples

    if os.path.isfile(csv_output_name) == True:
        print(f"Merged sample ID and barcode name file for pool {pool_num}, {csv_output_name}, has been created.\n")

def get_rawread_filename(pool_num):
    filename_regex = re.compile(('^(\d{4}\.Pool' + str(pool_num) + ')\S+(\.fastq\.gz)$'))
    rawread_file = list(filter(filename_regex.search, os.listdir(raw_read_path)))
    assert len(rawread_file) == 1, f'{len(rawread_file)} .fastq.gz files were located in the ' \
                                   f'provided filepath: {rawread_file} for pool {pool_num}. Requires exactly 1 ' \
                                   f'file per pool.'
    return raw_read_path + rawread_file[0]

def execute_lima_demux(pool_num):
    '''
    Composes the command line script for Lima demultiplexing, and executes it.
    :param run_prefix: four-digit prefix on sequencing run files, provided by sequencer
    :param pool_num: sequencing pool number, if split into more than one pool
    :return:
    '''
    start_runtime()
    rawread_file = get_rawread_filename(pool_num)
    barcode_file = barcode_path + run_name + "pool" + str(pool_num) + "_barcodes.fasta"
    output = barcode_path + run_name + "pool" + str(pool_num) + "_demux.fasta"

    lima_cmd = "lima " + rawread_file + " " + barcode_file + " " + output + \
               " --hifi-preset ASYMMETRIC" # asymmetric means forward and reverse barcodes are unique

    os.system(lima_cmd)
    end_runtime()
    print_runtime(custom_text=f'Lima demultiplexed samples from pool {pool_num}')

def get_barcode_names(pool_num):
    '''
    Get the name of the barcodes from the header of the unique
    barcode fasta files used in Lima demultiplexing.
    :param pool_num: the number of the PacBio sequencing pool (1/2)
    :return: list of the name of the barcodes
    '''
    barcode_names = []

    barcode_file = f"{output_path}{run_name}pool{str(pool_num)}_barcodes.fasta"

    with open(barcode_file) as fin:
        for line in fin:
            if line.startswith(">"):
                barcode_names.append(line.strip(">").strip("\n"))
    fin.close()

    return barcode_names

def create_post_lima_df(pool_num):
    '''
    Identifies barcode pairs based on index values provided by Lima, reorients the barcodes,
    and returns a dataframe of the read IDs, their sequences, and their barcodes.
    :param pool_num: sequencing pool number
    :return: dataframe of read IDs and their barcode combinations, ready for demultiplexing
    '''
    pool_num = str(pool_num)
    print(f"\nCreating a dataframe of the Lima output for pool {pool_num}...")
    start_runtime()
    barcode_list = get_barcode_names(pool_num) # get list of barcodes

    seq_sample_id = []  # read IDs
    bc_index_01 = []  # first barcode detected by Lima (returns index value)
    bc_index_02 = []  # second barcode detected by Lima (returns index value)
    seq = []  # sequence belonging to read ID, with barcodes removed by Lima

    # output of script 02_run-lima.py
    demux_fasta = f"{output_path}{run_name}pool{pool_num}_demux.fasta"

    with open(demux_fasta) as fin:
        for line in fin.readlines():
            if line.startswith(">"):
                split_line = line.split(" ")
                seq_sample_id.append(split_line[0][1:])
                bc_index_01.append(split_line[1].split(',')[0][3:])
                bc_index_02.append(split_line[1].split(',')[1])
            else:
                seq.append(line)
    fin.close()

    post_lima = pd.DataFrame({'seq_sample_id': seq_sample_id, # see notes for lists above
                              'bc_index_01': bc_index_01,
                              'bc_index_02': bc_index_02,
                              'dna_seq': seq})

    post_lima['bc_01_id'] = None  # create empty column to add barcode ID to
    post_lima['bc_02_id'] = None

    for i in range(post_lima.shape[0]):
        index_01 = post_lima['bc_index_01'][i]  # get the index of the barcode in the fasta file given to Lima
        index_02 = post_lima['bc_index_02'][i]
        for j in barcode_list:  # for each name in the barcode list (headers in fasta file given to Lima)
            bc_end = j.split("_")[2]  # only take the index (last part of the header string)
            if bc_end == index_01:  # if this index matches Lima's index assignment
                post_lima['bc_01_id'][i] = j  # add the full barcode name to the 'fwd' index id column
            if bc_end == index_02:  # if it makes the second index...
                post_lima['bc_02_id'][i] = j  # add to second index column (may be both sometimes, so two if statements, no elif)

    # some barcodes are oriented incorrectly (e.g., first barcode is reverse, second is forward)
    # to complete demultiplexing, need to reorient them, into new columns in df
    post_lima['bc_fwd_id'] = None
    post_lima['bc_rev_id'] = None
    post_lima['bc_index_fwd'] = None
    post_lima['bc_index_rev'] = None

    # reorient barcodes
    mismatch_count = 0
    for i in range(post_lima.shape[0]):
        bc_id_01 = post_lima['bc_01_id'][i]
        bc_id_01_dir = bc_id_01.split("_")[1]
        bc_id_02 = post_lima['bc_02_id'][i]
        bc_id_02_dir = bc_id_02.split("_")[1]
        if bc_id_01_dir == bc_id_02_dir:  # remove the ones that have two fwd or two rev barcodes
            mismatch_count += 1
            post_lima['bc_fwd_id'][i] = "NA"
            post_lima['bc_rev_id'][i] = "NA"
        else:
            if bc_id_01.split("_")[1] == "fwd":
                post_lima['bc_fwd_id'][i] = bc_id_01
                post_lima['bc_index_fwd'][i] = bc_id_01.split("_")[2]
                post_lima['bc_rev_id'][i] = bc_id_02
                post_lima['bc_index_rev'][i] = bc_id_02.split("_")[2]
            else:
                post_lima['bc_fwd_id'][i] = bc_id_02
                post_lima['bc_index_fwd'][i] = bc_id_02.split("_")[2]
                post_lima['bc_rev_id'][i] = bc_id_01
                post_lima['bc_index_rev'][i] = bc_id_01.split("_")[2]

    end_runtime()
    print_runtime(custom_text=f"The post-Lima dataframe for pool {pool_num} was created")

    return post_lima

# import the counts post-lima info to confirm that all reads were loaded into post_lima df
def check_counts(post_lima_df, pool_num):
    pool_num = str(pool_num)
    demux_lima_counts = f"{output_path}{run_name}pool{pool_num}_demux.lima.counts"
    post_lima_counts = pd.read_csv(demux_lima_counts, sep='\t')['Counts'].sum()
    if (post_lima_counts == post_lima_df.shape[0]):
        print(f"\nAll reads are loaded into the post lima dataframe for pool {pool_num}.")
    else:
        print(f"\n**ERROR**: Not all reads are loaded into the post lima dataframe for "
              f"pooll {pool_num}. Exiting...\n")
        sys.exit()

def create_pre_lima_df(pool_num):
    pool_num = str(pool_num)

    print(f"\nCreating a dataframe of the Lima input barcodes for {pool_num}, which allows "
          f"use to reorient barcode combinations that may be flipped...")
    start_runtime()

    merged_csv = f"{output_path}{run_name}pool{pool_num}_merged.csv"
    merged_df = pd.read_csv(merged_csv)
    merged_df.rename(columns = {'fwd_index': 'bc_index_fwd',
                                'rev_index': 'bc_index_rev'}, inplace=True)

    end_runtime()
    print_runtime(custom_text=f"The pre-Lima dataframe for pool {pool_num} was created")

    return merged_df

def get_bc_combo(df): # fwd and rev index columns must be named 'bc_index_fwd'
    '''
    Make a combination of the forward and reverse barcodes using the index number of the barcodes
    given to Lima in a fasta file, required in order to merge sample IDs with Lima output.
    :param df: dataframe containing barcodes
    :return: none, will add column to the input dataframe
    '''
    df['bc_combo'] = [ str(df['bc_index_fwd'][i]) + "_" + str(df['bc_index_rev'][i]) for i in range(df.shape[0])]
    return None

def final_demux(pre_lima_df, post_lima_df, pool_num):
    pool_num = str(pool_num)
    print(f"\nDemultiplexing samples in pool {pool_num}...")
    start_runtime()

    lima_merge_df = pre_lima_df.merge(post_lima_df, how='inner', on='bc_combo')

    sample_ids = set(lima_merge_df['sample_id'])
    read_for_count = []
    sample_for_count = []

    for i,s in enumerate(sample_ids):
        print(f"\r    sample {s} ({i+1}/{len(sample_ids)})--------", end='')
        sample_fasta_name = f"{final_demux_path}{run_name}{s}.fasta"
        read_count = 0
        for i in range(lima_merge_df.shape[0]):
            if lima_merge_df['sample_id'][i] == s:
                read_count += 1
                if os.path.isfile(sample_fasta_name) == True:
                    file_mode = "a"
                else:
                    file_mode = "w"
                with open(sample_fasta_name, file_mode) as fasta_out:
                    fasta_out.write(">" + lima_merge_df['seq_sample_id'][i] + "\n")
                    fasta_out.write(lima_merge_df['dna_seq'][i])
        read_for_count.append(read_count)
        sample_for_count.append(s)

    count_summary = pd.DataFrame({'seq_pool': pool_num,
                                  'sample_id': sample_for_count,
                                  'read_count': read_for_count})

    count_summary.to_csv(f"{final_demux_path}{run_name}pool{pool_num}_read-counts.csv", index=False)
    end_runtime()
    print_runtime(custom_text=f'\nSamples in pool {pool_num} have been demultiplexed')

# check which samples didn't make it to this final pool post-processing
def get_missing_samples(pool_num):
    '''
    Check which samples were missing from the final demultiplexing for a
    given quality score folder. Will provide a print out of the missing
    samples, and also saves a table of the missing samples.
    :param pool_num: number of the sequencing pool
    :return: saves a text file of missing samples in the pool in the
    final-demux folder
    '''
    pool_num = str(pool_num)

    print(f"\nChecking which (if any) samples are excluded from the Q{quality_score} "
          f"reads from pool {pool_num} (due to quality filtering)...\n")

    input_samples = pd.read_csv(f"{output_path}{run_name}pool{pool_num}_merged.csv")['sample_id'].tolist()
    output_samples = pd.read_csv(f"{final_demux_path}{run_name}pool{pool_num}_read-counts.csv")['sample_id'].tolist()

    missing_samples = set(input_samples).difference(set(output_samples))

    if len(missing_samples) > 0:
        print(f"    The following samples are missing from the Q{quality_score} samples in pool {pool_num}:")
        for sample in missing_samples:
            print(f"        - {sample}")

    missing_samples_out = f"{output_path}{run_name}pool{pool_num}_missing-samples.txt"
    with open(missing_samples_out, 'wt') as fout:
        for sample in missing_samples:
            fout.write(f"sample\n")
    fout.close()

    print(f"\n    The missing sample IDs have been saved to: {missing_samples_out}")

def execute_cutadapt(demux_fasta):
    demux_filepath = final_demux_path + demux_fasta
    trim_output = f"{output_path}trim_{demux_fasta}"
    untrim_output = f"{notrim_output_path}untrimmed_{demux_fasta}"

    cutadapt_cmd = f"cutadapt -g ^{fwd_primer}...{rev_revcomp_primer} -g ^{rev_primer}...{fwd_revcomp_primer} " \
                   f"-n 2 --quiet --untrimmed-output={untrim_output} " \
                   f"-o {trim_output} {demux_filepath}"

    os.system(cutadapt_cmd)

def check_read():
    '''
    Checks to see if a read (string of sequences from a single read)
    contains the forward and/or reverse primer in their forward, reverse,
    and reverse complement orientation.
    :return: prints error if a primer is detected
    '''
    with open((output_path + file), 'rt') as fin:
        lines = fin.readlines()
        for l,line in enumerate(lines):
            if line.startswith('>'):
                continue
            else:
                read = line.strip()
                for p,name in zip(primers, primer_names):
                    result = re.search(p, read)
                    if result is not None:
                        global primer_detect_count
                        primer_detect_count += 1
                        with open(primer_check_summary, 'a') as fout:
                            fout.write(f"\nSAMPLE {sample_id}____________________________\n")
                            fout.write(f"The {name} ({p}) was found starting at position {result.span()[0]} in read {lines[l-1]}")

# INCOMPLETE FUNCTION
# def get_read_lens(fasta_path):
#     '''
#     Get a dictionary of the sample Id as a key and
#     the length of each read belonging to that sample
#     as the values, as a list.
#     :param fasta_path: path to fasta files to get the
#     read lengths from
#     :return: dictionary of sample IDs and read lengths
#     '''
#     sample_ids = get_sample_id(fasta_path)
#
#     output_dict = {x:[] for x in sample_ids}
#
#     #for sample in sample_ids:
#     return None

def vsearch_dereplicate(trimmed_file):
    sample_id = trimmed_file.split('_')[-1].split('.')[0]

    input = trimmed_path + trimmed_file
    output_derep = f"{derep_path}{trimmed_file.replace('trim','derep')}"
    output_summary = f"{derep_summary_path}{trimmed_file.replace('trim','derep-summary').replace('fasta','txt')}"
    vsearch_derep_cmd = "vsearch --derep_fulllength " + input +\
        " --output " + output_derep + " --sizeout " +\
        " --uc " + output_summary + " --quiet"
    print(f"\r    sample {sample_id} ({f + 1}/{len(trimmed_files)})--------", end='')
    os.system(vsearch_derep_cmd)

def get_full_derep_summary(file_path_list):

    sample_ids = []
    num_unique_reads = []
    num_total_reads = []
    mean_cluster_size = []
    std_cluster_size = []
    min_cluster_size = []
    max_cluster_size = []

    for file in file_path_list:
        summary1 = pd.read_csv(file, delimiter="\t",
                              names=['record_type','cluster','cluster_size',
                                     'percent_centroid_similarity','match_orientation',
                                     'NA1','NA2','NA3','query_seq','centroid_seq']).drop(['NA1','NA2','NA3'], axis = 1)
        summary = summary1[(summary1['record_type'] != 'C')]
        sample_ids.append(file.split("/")[2].strip(".txt").replace('derep-summary_',""))
        num_unique_reads.append(len(summary['cluster'].unique()))
        num_total_reads.append(summary.shape[0])
        mean_cluster_size.append(round(np.mean(summary['cluster_size']),2))
        std_cluster_size.append(round(np.std(summary['cluster_size']),2))
        min_cluster_size.append(np.min(summary['cluster_size']))
        max_cluster_size.append(np.max(summary['cluster_size']))

    summary_df = pd.DataFrame(list(zip(sample_ids, num_unique_reads, num_total_reads, mean_cluster_size, std_cluster_size,
                                       min_cluster_size, max_cluster_size)),
                              columns=['sample_id','num_unique_reads','num_total_reads','mean_cluster_size',
                                       'std_cluster_size','min_cluster_size','max_cluster_size'])

    summary_df['percent_duplicated'] = round(((1 - (summary_df['num_unique_reads'] / summary_df['num_total_reads']))*100),2)

    summary_df.to_csv(df_out, index = False)

    return None

def vsearch_chimeras(derep_file):
    sample_id = derep_file.split('_')[-1].split('.')[0]
    input = derep_path + derep_file
    output_nochim = f"{nonchim_path}{derep_file.replace('derep','nochim')}"
    output_chim = f"{chim_path}{derep_file.replace('derep','chim')}"
    output_summary = f"{chim_summary_path}{derep_file.replace('derep','uchime-summary').replace('fasta','txt')}"
    vsearch_chimera_cmd = f'vsearch --uchime_denovo {input} --chimeras {output_chim} ' \
                          f'--nonchimeras {output_nochim} --uchimeout {output_summary} ' \
                          f'--quiet'
    print(f"\r    sample {sample_id} ({f + 1}/{len(derep_files)})--------", end='')
    os.system(vsearch_chimera_cmd)

def get_size(header):
    size = header.split(";")[-1].replace("size=","")
    return int(size)

def vsearch_filter(nonchim_file):
    sample_id = nonchim_file.split('_')[-1].split('.')[0]
    input = nonchim_path + nonchim_file
    output_filter = f"{filtered_path}{nonchim_file.replace('nochim','filt')}"
    discarded_filter = f"{discard_path}{nonchim_file.replace('nochim','discard')}"
    vsearch_filter_cmd = f"vsearch --fastx_filter {input} --fastaout {output_filter} " \
                         f"--minsize {minsize_threshold} --quiet --fastaout_discarded {discarded_filter}"
    print(f"\r    sample {sample_id} ({f + 1}/{len(nonchim_files)})--------", end='')
    os.system(vsearch_filter_cmd)

def mothur_align_all(filtered_file):
    sample_id = filtered_file.split('_')[-1].split('.')[0]
    input = f"{filtered_path}{filtered_file}"  # if run from CLI, no need to escape '-' characters

    namefile_cmd = f"unique.seqs(fasta={input}, outputdir={count_output_path})"
    align_cmd = f"pairwise.seqs(fasta={input}, cutoff={cutoff}, countends=F, outputdir={distmat_output_path}, processors={num_processors})"
    align_log_cmd = f"set.logfile(name={distmat_log}, append=T)"
    names_log_cmd = f"set.logfile(name={count_log}, append=T)"

    print(f"\r    sample {sample_id} ({f + 1}/{len(filtered_files)})--------", end='')
    mothur_run_cmd = f"{mothur_path} '#{align_log_cmd};{align_cmd};{names_log_cmd};{namefile_cmd};' -q" # no matter what i do it won't be quiet
    os.system(mothur_run_cmd)

def mothur_cluster(distmat_file):
    sample_id = distmat_file.split('_')[-1].split('.')[0]
    distmat_input = f"{distmat_path}{distmat_file}"
    count_input = f"{count_path}{distmat_file.replace('dist','count_table')}"

    cluster_log_cmd = f"set.logfile(name={cluster_log}, append=T)"
    mothur_cluster_cmd = f"{mothur_path} '#{cluster_log_cmd};cluster(column={distmat_input}, " \
                         f"outputdir={cluster_path}, count={count_input}, cutoff={cutoff}, " \
                         f"method=average)'"
    print(f"\r    sample {sample_id} ({f + 1}/{len(distmat_files)})--------", end='')
    os.system(mothur_cluster_cmd)

def cast_lab_as_float(input_table):
    input_table02 = input_table[input_table['label'] != 'unique']
    input_table02['label'] = input_table02['label'].astype(dtype=float)
    return input_table02



from pathlib import Path

NEEDS_ACTION_REGEX = r'needs_(?!demux)'

# PACBIO_DIR = 'pacbio'
# ILLUMINA_DIR = 'illumina'
# SANGER_DIR = 'sanger'

# PLATFORM LIST
SEQ_PLATFORM_OPTS = ['pacbio', 'illumina', 'sanger']

# CONFIGURATION FILE HANDLES
CONFIG_FILETYPE = '.toml'
RENAME_CONFIG_HANDLE = 'file-rename'
PIPELINE_CONFIG_HANDLE = 'pipeline-settings'
NAME_CONFIG_HANDLE = 'filename-components'

# CONFIGURATION FILE SECTIONS
PIPELINE_CONFIG_MPID = 'pacbio-multiplex-ids'

# MISCELL REGEX
AFFIRM_REGEX = '^c|(continue)|^y'

YES_RE = '^yes$|^y$'
NO_RE = '^no$|^n$'
QUIT_RE = '^quit.*?|^exit.*?'  # avoids confusion about whether to use quit or quit() or exit or exit()

HIDDEN_FILE_REGEX = '^\.'
GZIP_REGEX = '(\.fast.\.gz)$'
ANY_PLATFORM_REGEX = '^illumina|^pacbio|^sanger'
SEQ_FILE_GLOB = '*.fast*'

# PIPELINE FILE PREFIXES
NOPHIX_PREFIX = 'no-phix'
NOAMBIG_PREFIX = 'no-ambig'
TRIMMED_PREFIX = 'trim'
DEREP_PREFIX = 'derep'
MERGED_PREFIX = 'merged'
QUALFILT_PREFIX = 'qualfilt'
ITSX_PREFIX = 'itsx'
NOCHIM_PREFIX = 'no-chim'
CLUSTER_PREFIX = 'clust'
REPREAD_PREFIX = 'rep-read'

# PIPELINE FILE SUFFIXES
LOG_SUFFIX = '.log'


import warnings
import pandas as pd
import matplotlib.pyplot as plt

# set user preferences for warnings, figure quality, etc.
pd.options.mode.chained_assignment = None  # default='warn', blocks copy on df warning
warnings.simplefilter(action='ignore', category=FutureWarning)  # supress future warnings
pd.set_option('display.max_columns', 3000)
plt.rcParams['savefig.dpi'] = 300



import subprocess, re, sys, pathlib, tomlkit
from pathlib import Path
from datetime import datetime
from functools import wraps
import shutil
import pandas as pd
from climush.constants import *

# tried to order by category but some required specific order due to dependencies on other functions

#######################
# MISC UTILITIES ######
#######################

# check if the provided file path is a pathlib Path object
def is_pathclass(file_path, exit_if_false=True):
    '''
    Checks if the input is a Path class.

    :param file_path: path in question
    :param exit_if_false: whether to exit the script if the
    file path is found to not be a valid Path object; defaults
    to true, as it tends to mess up downstream processes.
    :return: True/error message
    '''
    if isinstance(file_path, pathlib.PurePath):
        return True
    else:
        msg = f'The provided path, {file_path}, is not a valid Path object.\n'
        print(msg)
        if exit_if_false:
            return exit_process(message=msg)
        else:
            return False

# alert if multiple files matching the search pattern in the file path are returned
# prompt for user input if multiple or no files matching the pattern are returned
# moved here because used in continue_to_next()
def flag_multiple_files(file_path, search_for):
    assert is_pathclass(file_path, exit_if_false=False)

    result = list(file_path.glob(search_for))

    if len(result) == 1:
        return result[0]
    elif len(result) == 0:
        all_files = file_path.glob('*')
        print(f'No files matching the pattern \'{search_for}\' were detected in the file path: {file_path}. Do '
              f'you mean any of these files in this directory?')
        which_file = prompt_print_options([all_files, 'none of these (exit)'])
        if which_file in all_files:
            return which_file
        else:
            return exit_process(message=f'The response \'none of these\' was chosen when searching for the correct'
                                        f'file matching the pattern: {search_for}')
    else:
        print(f'{len(result)} files matching the pattern \'{search_for}\' were detected in the file path: '
              f'{file_path}. Please type the number corresponding to the correct file to use:')
        which_file = prompt_print_options([result, 'none of these (exit)'])
        if which_file == 'none of these (exit)':
            return exit_process(message=f'The response \'none of these\' was chosen when searching for the correct'
                                        f'file matching the pattern: {search_for}.')
        else:
            return which_file

# activate next script in python
def continue_to_next(this_script, config_dict):
    '''
    Continue to the next step of the pipeline.

    Will activate the script from the CLI with default
    parameters. If you do not want to use the default
    parameters, opt out of automated continuation by
    responding to the CLI prompt that precedes the use
    of this function.
    :param this_script: use __file__ always
    :return: None, assembles and runs a shell command.
    '''
    automate_dict = config_dict['automate']
    auto = automate_dict['run_all']
    to_run = automate_dict['run_some']
    to_exclude = automate_dict['exclude']

    current_script = Path(this_script).stem
    current_num = int(current_script.split('_')[0])

    def format_num_for_search(current_script_number):
        if current_script_number > 9:
            next_num_search = str(current_script_number + 1)
        else:
            next_num_search = '0' + str(current_script_number + 1)
        return next_num_search

    next_num_search = format_num_for_search(current_num)
    next_script = flag_multiple_files(file_path=this_script.parent, search_for=f'{next_num_search}*')

    if (auto) or ((current_num + 1) in to_run):
        print(f'\n\nRunning next step, {next_script.name}...\n')
        subprocess.run(['python3', next_script])
        return None
    elif (current_num + 1) in to_exclude:
        i = 2
        while (current_num + i) in to_exclude:
            i += 1
        next_num_search = format_num_for_search(current_num + 1)
        next_wanted = flag_multiple_files(file_path=this_script.parent, search_for=f'{next_num_search}')
        print(f'\n\nRunning next step, {next_wanted.name}...\n')
        subprocess.run(['python3'], next_wanted)
        return None
    else:
        to_next = input(f'The script {current_script} has completed. Would you like to continue '
                        f'to the next step in the pipeline, {next_script.stem}?[Y/N]\n'
                        f'\n')
        if re.search(AFFIRM_REGEX, to_next, re.I):
            print(f'\n\nRunning next step, {next_script.name}...\n')
            subprocess.run(['python3', next_script])
            return None
        else:
            return print(f'\n\nExiting the completed step, {current_script}.\n')

# exit current script due to error, save script and timestamp of where error occurred
def exit_process(message, config_section='error.message'):
    script_name = sys.argv[0]  # unsure if will get name of script it is executed in or the one it is compiled in
    exit_time = datetime.today().strftime('%Y-%m-%d %H:%M:%S')
    # write_to_config(config_section=config_section, script=script_name, timestamp=exit_time, details=message)
    print(f'Exiting {script_name}...\n')
    return sys.exit()

# recursive function that will seek out yes/no/quit response continuously until achieved
def prompt_yes_no_quit(message):
    '''
    User prompt accepting specific responses.

    Takes the input of a print message that shows just prior to a user prompt. The message
    should specify the potential responses of yes/y, no/n, or quit/exit (case insensitive).
    If the user inputs a response that is not one of these potential responses, it will provide
    a notification that the input response was not recognize, and relist the acceptable responses.
    It then will immediately reprint the initial prompt to allow the user another chance to enter
    and acceptable input string.
    :param message: any message describing what is wanted from the user; see the return for how
    to set up the message so that the desired output is achieved
    :return: if response is yes/y, it will return nothing and then next line of code in the
    script will be run; if response is no/n/quit/exit, then it will exit the current script; if the
    response is unfamiliar, it will print an alert for invalid input, and restart the prompt
    again until an acceptable answer is received.
    '''

    print(f'{message} [yes/no/quit]\n')

    response = input()

    if re.search(YES_RE, response, re.I):
        return None
    elif re.search(NO_RE, response, re.I) or re.search(QUIT_RE, response, re.I):
        print(f'Exiting...\n')
        return sys.exit()
    else:
        print(f'Your response was not recognized out of the list of potential responses. Please respond with '
              f'\'yes\', \'no\', or \'quit\'.\n')
        return prompt_yes_no_quit(message)

# print out list of options for option-based prompts
def prompt_print_options(option_list):
    option_list.sort()  # sort option list for easier user experience
    option_list.append('quit')  # add option to quit, listed last

    for o,opt in enumerate(option_list):
        if (o+1) < len(option_list):
            print(f'\t[{o+1}]: {opt}')  # print out all options from option list, excluding final option
        else:
            selected_opt = input(f'\t[{o+1}]: {opt}\n')  # print out last option from option list as prompt

    result = option_list[int(selected_opt)-1]  # return the item from option list matching the user input

    if result == 'quit':
        return sys.exit()
    else:
        return result

# print a CLI prompt when multiple files are detected when only one is expected
def prompt_multiple_files(file_path):
    print(f'\nWARNING: Multiple files were detected in the '
          f'{file_path.stem} folder. Please type the number '
          f'corresponding to the correct file to use:')
    file_list = [file.stem for file in file_path.glob('*') if not re.search(HIDDEN_FILE_REGEX, file.stem)]
    return prompt_print_options(file_list)

# provide options for the sequencing platform if it cannot be otherwise detected
def prompt_sequencing_platform(sample_id):
    print(f'\nWARNING: The sequencing platform could not be inferred from the sample: {sample_id}. Please type '
          f'the number corresponding to the correct sequencing platform from the options below. ')
    platform = prompt_print_options([SEQ_PLATFORM_OPTS, 'multiple'])
    if platform == 'multiple':
        print(f'If you have a combination of sequences from multiple platforms, you will need to either:\n'
              f'(1) manually sort the files into their sequencing platform directories\n'
              f'(2) use the file renaming script, rename_sequence_files.py, to rename the files to match '
              f'the file naming convention.\n')
        return

# print indented list
def print_indented_list(print_list):
    print_list[0] = '\t' + print_list[0] # add leading tab for first item printed from list, otherwise adds after line break
    formatted_list = '\n\t'.join(print_list)
    return print(formatted_list)

# run shell command and save stdout and stderr to file
def run_subprocess(cli_command_list, dest_dir):
    '''
    Run a subprocess, saving the output and error to a log file.

    Takes the command line (CL) argument assembled as a list, and runs
    the subprocess, saving the standard in (stdin), standard out (stdout),
    and standard error (stderr) from the process. The output of stderr and
    stdout are written to a log file in the destination directory
    :param cli_command_list: the list of the components of the command line
    argument to run through subprocess; each argument should be separated
    out into a list, rather than a single string.
    :param dest_path: directory to write the file to; should be a Path
    object
    :return: None
    '''
    assert is_pathclass(dest_dir)

    assert isinstance(cli_command_list, list) and len(cli_command_list) > 1, print(f'Command line process to run '
                                                                             f'should be in a list of its components. '
                                                                             f'Try shlex.split() if unsure how to '
                                                                             f'compose the list. ')

    # run_cmd = subprocess.run(cli_command_list, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    # out, err = run_cmd.communicate()
    program = cli_command_list[0]

    if re.search('.\..', program):
        program = program.split('.')[0]

    run_cmd = subprocess.run(cli_command_list, capture_output=True)

    out_path = dest_dir / f'{program}.out'
    with open(out_path, 'at') as fout:
        as_str = run_cmd.stdout.decode('utf-8')
        if int(out_path.stat().st_size) == 0:
            fout.write(as_str)
        else:
            fout.write(as_str.split('\n')[1] + '\n')

    err_path = dest_dir / f'{program}.err'
    with open(err_path, 'ab') as fout:
        fout.write(run_cmd.stderr)

    temp_file = dest_dir / f'{program}.temp'

    if len(run_cmd.stderr) == 0:
        pass
    else:
        if not temp_file.is_file():
            continue_ok = input(f'Running {program} produced an error. Please review the output in {err_path.name}. '
                                f'Would like to continue despite this error? [Y/N]')
            if re.search(continue_ok, AFFIRM_REGEX, re.I):
                with open(temp_file, 'wt') as fout:
                    fout.write('stop_prompt')
                return None
            else:
                return exit_process(message=f'Running {program} produced an error. See {err_path.name} for details.')

    return None

def func_timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = datetime.datetime.now()
        func_output = func(*args, **kwargs)
        end_time = datetime.datetime.now()
        runtime = str(end_time - start_time).split('.')[0]  # round seconds down
        print(f"{func.__name__} was executed in {runtime}.\n")
        return func_output
    return wrapper()

def get_seq_platform(fastx_file, delim):
    if isinstance(fastx_file, str):
        sample_id = fastx_file
    elif is_pathclass(fastx_file):
        sample_id = fastx_file.stem
    else:
        print(f'ERROR. An error occurred when trying to detect the sequencing platform from the '
              f'sample ID: {fastx_file}\n')


    platform = sample_id.split(delim)[0]
    if platform in SEQ_PLATFORM_OPTS:
        return platform
    else:
        pass

def import_mapping_df(df_path):
    '''
    Import .csv, .txt, or .xlsx table as a dictionary.

    Read in a mapping file for demultiplexing. Can accomodate the file formats
    .xlsx, .csv, and .txt. Will output a dictionary, where the key is the name
    of the tab in the dataframe and the value is the dataframe in that tab. Will
    always return a dictionary, although only .xlsx files will have tabs. Formats
    the name of the keys in the output dictionary to be 'pool1' or 'pool2'; the
    number of the pool is inferred from the name of the tab (.xlsx) or the name
    of the file (.csv, .txt). Returned as dictionary so that the output can be
    handled in the same way, regardless of file type (e.g., loop through tabs
    even if a .csv, which will have a single tab).
    :param df_path: path to the dataframe
    :return: dictionary, where the key is 'pool1' or 'pool2', and the value is the
    dataframe belonging to that tab, or in the case of a .csv or .txt file, the
    entirety of that dataframe file.
    '''

    POOL_NUM_RE = '(?<=pool).?(\d)'

    if re.search('^\.x', df_path.suffix):  # if an excel file
        mapping_tabs = pd.read_excel(df_path, sheet_name=None)  # need to set sheet_name to None to get all tabs
        old_tab_names = list(mapping_tabs.keys())  # make list of old names, otherwise cannot update key names in loop
        for tab in old_tab_names:
            try:  # format the name of the tab to be uniform across all tabs/dataframes
                pool_num = re.search(POOL_NUM_RE, tab, re.I).group(0)
            except AttributeError:
                print(f'The pool number could not be inferred from tab {tab} in the mapping file {df_path.name}. '
                      f'Please type the correct pool number for this file: ')
                pool_num = prompt_print_options(['1', '2'])  # choose from 1 or 2 (or quit, built into function)
            mapping_tabs[f'pool{pool_num}'] = mapping_tabs.pop(tab)  # replace old tab (key) with reformatted one
    elif re.search('^\.c|^\.txt$', df_path.suffix):  # I think you can read in .txt and .csv files the same way?
        try:  # try to get the pool number from the file name
            pool_num = re.search(POOL_NUM_RE, df_path.name, re.I).group(0)
        except AttributeError:  # if there's no detected pool number in the name, prompt user to specify one
            print(f'The pool number could not be inferred from the file name of the mapping file {df_path.name}. '
                  f'Please type the correct pool number for this file: ')
            pool_num = prompt_print_options(['1', '2'])  # choose from 1 or 2 (or quit, built into function)
        mapping_tabs = {f'pool{pool_num}': pd.read_csv(df_path)}  # make dict to match format from .xlsx
    else:
        print(f'ERROR. The file format {df_path.suffix} of the mapping file {df_path.name} is not a recognized '
              f'file type. Accepted file types are: \'.xlsx\', \'.csv\', and \'.txt\'.\n')
        sys.exit()

    return mapping_tabs


#######################
# FILE PATHS ##########
#######################

def filter_empty_files(dir_path, keep_empty=True):
    empty_files = 0
    empty_folders = 0
    for file in dir_path.glob('*'):
        if file.stat().st_size == 0:
            empty_files += 1
            if keep_empty:
                # REPLACE W/ MKDIR EXIST OKAY
                dest_dir = dir_path / 'empty-files'
                dest_dir.mkdir(exist_ok=True)
                dest_path = dest_dir / file.name
                file.rename(dest_path)
                fate = f'moved to the directory {dest_dir}'
            else:
                file.unlink()
                fate = 'deleted'
        elif file.is_dir():
            try:  # empty dirs not always size 0 (hidden files); this will only rm dir if it is empty (but not 0)
                file.rmdir()
                fate = 'deleted'
                empty_folders += 1
            except:
                continue
                fate = ''
        else:
            continue
            fate = ''

    return print(f'{empty_files} files and {empty_folders} folders were {fate}.\n')

# check whether a path exists; return path if it does or raise error if it does not
def flag_if_not_path_exists(file_path, absolute=True, exit_if_false=True):
    '''
    Create a Path class using the provided filepath string.

    If the filepath does not exist, send an error message,
    and will exit the script.
    :param filepath: string of absolute or relative path.
    :param absolute: default True; if True, will return the
    Path as an absolute path; otherwise will remain relative
    (which may cause issues when joining paths later on)
    :param exit_if_false: whether to exit the Python script if
    the provided path does not exist; defaults to True
    :return: filepath as PosixPath if it exists, otherwise
    an error will exit the script.
    '''
    try_path = Path(file_path)

    if try_path.exists():
        if absolute:
            return try_path.resolve()
        else:
            return try_path
    else:
        msg = f'Cannot locate the filepath {file_path}. Please '\
              f'make sure that either the original file structure '\
              f'of the container is unchanged, or use the --seq-path/-sp '\
              f'flags while running this script.\n'
        print(msg)
        if exit_if_false:
            return exit_process(message=msg)
        else:
            return None

# create a log file to write details of pipeline step to
def make_log_file(file_name, dest_path, log_suffix=LOG_SUFFIX):
    '''
    Create file path for a log file.

    Files with multiple file extensions (e.g., fastq.gz) are
    not easily manipulated with most standard libraries like
    pathlib. Only '.gz' will be recognized as the file suffix/
    extension. This function deals with the different approaches
    required when the file extension is complex (.fastq.gz) and
    when it is simple (.fasta)
    :param file_name: name of the file to use to base
    the log file name off of
    :param log_suffix: the suffix/file extension to use for the log
    file; defaults to '.log'
    :param dest_path: directory to create the file
    :return: path to log file
    '''
    assert is_pathclass(file_name)
    assert is_pathclass(dest_path)

    if not log_suffix.startswith('.'):  # for user-provided suffix, check there is a '.'
        log_suffix = '.' + log_suffix

    if re.search(GZIP_REGEX, file_name.name):
        log_file_name = re.sub(GZIP_REGEX, log_suffix, file_name.name)
    else:
        log_file_name = file_name.with_suffix(log_suffix)

    return Path(dest_path / log_file_name)

# create directory if it doesn't exist and return path, or return the directory path if it already exists
def mkdir_exist_ok(new_dir, parent_dir=None):
    '''
    Make a new directory if one does not exist. Bypass making
    the directory if it does exist.
    :param new_dir: name of new directory as string
    :param parent_dir: path object to new directory; if none
    is specified, the new directory must be an absolute path.
    :return: Path object of directory path
    '''
    if not (isinstance(new_dir, str)) and (new_dir.is_absolute()):
        new_dir.mkdir(exist_ok=True)
        return new_dir
    elif (parent_dir is None):
        return print(f'\nThe provided file path of the new directory is not absolute. Please '
                           f'provide a parent_dir to specify the path of the new directory, and '
                           f'retry.\n')
    else:
        new_path = Path(parent_dir / new_dir)
        new_path.mkdir(exist_ok=True)
        return new_path

# move files from current location (source_path) to new location (dest_path)
def move_file(source_path, dest_path):
    assert isinstance(source_path, pathlib.PurePath), f'The provided source path is not a Path object: {source_path}'
    assert isinstance(dest_path, pathlib.PurePath), f'The provided destination path is not a Path object: {dest_path}'

    full_dest_path = dest_path / source_path.name

    source_path.rename(full_dest_path)

    if full_dest_path.is_file():
        return None
    else:
        return print(f'The file {source_path.name} was not properly copied to its new destination: {dest_path}.')

# sort the sequences provided in the 'sequences' input directory

# count the number of files in a directory
def count_files(file_path, search_for='*'):
    '''
    Counts the number of files in a file path, and returns as int.

    :param file_path: path to directory to count files in.
    :param search_for: the .glob string that describes the file types
    to search for; default is '*', which is all files
    :return: int of the number of files in directory
    '''
    # check if it is a file path that exists
    flag_if_not_path_exists(file_path, absolute=False, exit_if_false=False)

    return len(list(file_path.glob(search_for)))


def check_for_input(file_dir, seq_platform=None, file_ext=SEQ_FILE_GLOB):
    '''
    Checks if there are input files for the process.

    :param file_dir: the directory path to check for files and to check
    whether the path exists
    :param seq_platform: the type of sequencing files to look for ['illumina',
    'sanger', 'pacbio']; defaults to None, meaning it will return True if any
    non-empty directory or file is located
    :param file_ext: the expected file extension of the files to search for,
    using asterisk '*' for wildcard (used in glob)
    :return: two items; [1] Boolean T/F, whether there is a directory in this
    path that contains sequencing files; [2] a list of the files matching the
    file extension in the directory, if present
    '''
    assert is_pathclass(file_dir)

    # I decided to leave this out so you could put any regex in (i.e., multiplexed pacbio files will start with \d{4}
    # accepted_seq_input = [None, 'illumina', 'sanger', 'pacbio']
    # if not seq_platform in accepted_seq_input:
    #     print(f'The provided sequencing platform, {seq_platform}, is not one of the accepted '
    #           f'values: {accepted_seq_input}. Please retry with one input value from this list '
    #           f'as a string, or multiple values as a list of strings.\n')
    #     return sys.exit()

    if seq_platform is None:
        seq_re = '.+?'  # will return all files in directory
    elif isinstance(seq_platform, list):
        seq_re = '|'.join(seq_platform)  # returns only those in provided list
    else:
        seq_re = seq_platform  # returns only those of the provided platform

    if file_dir.is_dir():  # check that input is a directory
        if count_files(file_path=file_dir, search_for=file_ext) > 0:  # confirm it is not empty
            file_list = [f for f in file_dir.glob('*') if re.search(seq_re, f.name, re.I)]  # check for specific files
            if len(file_list) > 0:
                return True, file_list  # return true if files matching criteria are found
            else:
                print(f'The directory {file_dir} contains files, but none that match the platform-specific search '
                      f'criteria: {seq_platform}.')
                return False, file_list  # return false if not
        else:
            file_list = []  # if input is a directory, but it is empty...
            print(f'The directory {file_dir} exists, but is empty.\n')
            return False, file_list  # return false and empty list
    else:
        file_list = []  # if input is not a directory
        print(f'The directory {file_dir} does not exist.\n')
        return False, file_list  # return false and empty list

# get name of previous script

#######################
# TEXT MANIPULATION ###
#######################

# add prefix to file name
def add_prefix(file_path, prefix, dest_dir, action='rename', f_delim='_'):
    '''
    Add prefix to file name.

    Adds a prefix to the file name that indicates which step of the
    pipeline the given file is produced from. Typically, will use one
    of the constants for the prefix, but any string can be provided.
    Prefix constants only exist for the files that persist to the
    next step of the pipeline. To create prefixes for the non-persistent
    files (e.g., reads/samples that are filtered out), use the flip_prefix
    function from the textmanip module.
    :param file_path: either a directory containing files to add a prefix to
    or a single file to add a prefix to
    :param prefix: prefix to add to the file
    :param f_delim: file name separator to use between the prefix and the rest
    of the file name; defaults to an underscore
    :return: Path object with new file name
    '''

    assert is_pathclass(file_path, exit_if_false=False)

    def prefix_single(file_name):
        old_name = file_name.name
        platform_present = re.search('illumina|pacbio|sanger', old_name, re.I)
        try:
            # if there's a match, put prefix before the platform, remove all else
            location = platform_present.span()[0]
            new_name = prefix + f_delim + old_name[location:]
        except:
            # if there's no match, just add prefix to start of file name
            new_name = prefix + f_delim + old_name

        # replace old name with new name
        if action == 'rename':
            new_path = file_name.rename(dest_dir / new_name)
        elif action == 'copy':
            new_path = dest_dir / new_name
            shutil.copy(file_path, new_path)
        else:
            new_path = dest_dir / new_name

        return new_path


    if file_path.is_dir():
        new_path_list = []
        for file in file_path.glob('*'):
            new_path_list.append(prefix_single(file))
        return new_path_list
    elif file_path.is_file():
        return prefix_single(file_path)
    else:
        msg = f'FAILURE. Unrecognized input file; cannot recognize as either a directory or file.\n'
        return exit_process(msg)

    # filename_start = file_name.name.split('_')[0]
    # if re.search(ANY_PLATFORM_REGEX, filename_start, re.I):
    #     return dest_path / f'{prefix}_{file_name.name}'
    # elif re.search(ANY_PLATFORM_REGEX, file_name.name.split('_')[1], re.I):
    #     return dest_path / file_name.name.replace(filename_start, prefix)
    # else:
    #     msg = f'\nERROR: Cannot locate start of the original filename to add prefix to for '\
    #           f'file: {file_name}. Please make sure that the file name either starts with one '\
    #           f'of the prefixes expected given the naming convention (i.e., sanger, illumina, pacbio), '\
    #           f'or one of these expected prefixes is the second component of the filename.\n'
    #     print(msg)
    #     return exit_process(message=msg)
    #     return exit_process(message=msg)

def script_name_as_dir(script_name, parent, remove_num=False, suffix=None):
    '''
    Create a directory matching the name of the script from which it is executed.

    :param script_name: typically can just use __file__, but if I were to write that
    into the function, it would be the name of the module script for the climush
    Python package, not the file the function was used in
    :param suffix: suffix to add to end of directory name; may be useful
    to reflect the user-settings for that file
    :param parent: the parent directory under which to nest the directory; will
    default to the output directory of the pipeline, as that's what this function
    tends to be used for.
    :return: Path object to resulting directory
    '''
    assert is_pathclass(parent)

    if not parent.is_dir():
        parent = mkdir_exist_ok(new_dir=parent)

    script_num = Path(script_name).stem.split('_')[0]
    script_name_nonum = re.sub('^\d+_(?=\w)','',Path(script_name).stem)
    completed_name_nonum = OUTPUT_DIRS[script_name_nonum]
    completed_name = f'{script_num}_{completed_name_nonum}'

    if remove_num:
        completed_name = re.sub('^\d+_(?=\w)','', completed_name)

    if suffix:
        out_path = parent / f'{completed_name}_{suffix}'
    else:
        out_path = parent / completed_name

    out_path.mkdir(exist_ok=True)

    return out_path

def flip_prefix(prefix_const):
    # not sure why I would check if the prefix was a path? I don't think I ever use a path, just a constant string?
    # was throwing error when checking for PhiX in prefilter step, though was creating directory, error creating file
    # if is_pathclass(prefix_const, exit_if_false=False):
    #     prefix_const = prefix_const.stem
    # else:
    #     pass

    if re.search('^no-', prefix_const, re.I):  # if provided prefix starts with no, remove no
        return re.sub('^no-', '', prefix_const, re.I)
    else:  # if it does not start with no, add no
        return 'no-' + prefix_const

def rename_read_header(fasta_file, header_delim=';'):
    # fasta_file = next(post_itsx.glob('*.fast*'))
    parent_dir = fasta_file.parent.name

    if re.search('itsx', parent_dir, re.I):  # if they are post-itsx reads...

        READ_COUNT_RE = '(?<=size=)[0-9]{1,}'
        READ_LEN_RE = '[0-9]{1,}(?=\sbp)'
        READ_ID_RE = '[0-9]{1,}(?=\/ccs)'
        READ_REGION_RE = '(?<=\|\w\|).+(?=\sExtracted)'
        SAMPLE_ID_RE = '(?<=\w_)(pacbio|sanger|illumina).+?(?=\.)'

        # REPLACE WITH REMOVE PREFIX??
        sample_id = re.search(SAMPLE_ID_RE, fasta_file.name).group(0)

        updated_records = []
        no_update_count = 0
        for record in SeqIO.parse(fasta_file, 'fasta'):
            header = record.description
            try:
                read_id = re.search(READ_ID_RE, header).group(0)
                region = re.search(READ_REGION_RE, header).group(0)
                read_count = re.search(READ_COUNT_RE, header).group(0)
                read_length = re.search(READ_LEN_RE, header).group(0)

                updated_header = header_delim.join([f'{sample_id}_{read_id}', region, f'region_len={read_length}bp',
                                                    f'full-len_copies={read_count}'])

                record.id = updated_header
                record.name = f'{sample_id}_{read_id}'
                record.description = ''
                updated_records.append(record)
            except:
                no_update_count += 1
                updated_records.append(record)

        SeqIO.write(updated_records, fasta_file, 'fasta')

        if no_update_count == 0:
            return print(f'SUCCESS. All reads ({len(updated_records)}) were renamed '
                         f'in {fasta_file.name}.\n')
        elif no_update_count == len(updated_records):
            return print(f'FAILURE. None of the reads were renamed in {fasta_file.name}. This can '
                         f'happen if the fasta file has already been renamed by this function, '
                         f'so open the fasta file to check.\n')
        else:  # if some reads were renamed, but not all
            return print(f'FAILURE. {no_update_count} sequence headers could not be renamed due to '
                         f'a missing data field for reads in {fasta_file.name}.\n')
    else:
        return print(f'UNDER CONSTRUCTION. I haven\'t yet updated this function to work with any other '
                     f'fasta files than those produced after ITSx.\n')

def escape_path_special(file_path):
    if is_pathclass(file_path):
        file_path = str(file_path)

    if re.search('\(', file_path):
        file_path = re.sub('\(', '\(', file_path)

    if re.search('\)', file_path):
        file_path = re.sub('\)', '\)', file_path)

    return file_path

#######################
# CONFIGURATION #######
#######################

# define function to simplify import of configuration files
def import_config_as_dict(file_path, file_handle, config_section='all'):
    '''
    Import a configuration file and return as a dictionary.

    Checks the input path to ensure that the filepath exists; if
    not, it will stop the process and exit the entire script.
    Uses a custom parser to preserve the case of the text in the
    configuration file (i.e., uppercase letters remain uppercase,
    lowercase remain lowercase). Creates a dictionary where the
    sections of the configuration file are the primary keys, if it
    is a nested dictionary. If not a nested dictionary, will be
    returned as an un-nested dictionary where the keys are the
    section's keys.
    :param file_handle: standard file handle used for the config file; will
    default to the constants that work if the file structure of the container
    is maintained
    :param config_section: option to output only one section of the configuration
    file.
    :return: dictionary of values in the configuration file
    '''
    path_to_file_abs = flag_if_not_path_exists(file_path, absolute=True)

    config_file_list = list(path_to_file_abs.glob(f'*{file_handle}*'))

    if len(config_file_list) > 1:
        msg = f'\nERROR. Multiple configuration files matching the provided file handle, {file_handle}, were located '\
              f'in this directory:\n'\
              f'\t{[file.name for file in config_file_list]}\n'\
              f'Please include more information in the file handle input, then retry.\n'
        print(msg)
        # return exit_process(message=msg)
    else:
        config_file = config_file_list[0]

    with open(config_file, 'rt') as fin:
        config_dict = tomlkit.parse(fin.read())

    if config_section == 'all':
        return config_dict
    else:
        if config_section in config_dict.keys():
            return config_dict[config_section]
        else:
            msg = f'The provided configuration file section header, {config_section}, is not a valid header for ' \
                  f'the {file_handle} configuration file. Please enter the number of the correct header that you ' \
                  f'want:'
            print(msg)
            correct_header = prompt_print_options(list(config_dict.keys()))
            return config_dict[correct_header]

# write or update settings to the configuration file
# NOT WORKING RIGHT NOW
# def write_to_config(config_section, output_path, update_obj=False, overwrite=True, **kwargs):
#     # import the pipeline configuration file directly
#     config_section = 'error_message'
#     config_dict = import_config_as_dict(output_path, file_handle)
#
#     # confirm that the provided section is in the configuration file
#     if not config_section in config_dict.keys():
#         msg = f'ERROR: Could not locate the provided section \'{config_section}\' in the pipeline configuration '\
#               f'file.'
#         print(msg)
#         return exit_process(message=msg)
#
#     # update the configuration file with the provided key-value pairs
#     for k, v in kwargs.items():
#         if k in config_dict[config_section].keys():  # if there's already a key for this kwarg...
#             if (config_dict[config_section][k] == '') or overwrite:  # if it has no value or if we want to overwrite...
#                 config_dict[config_section][k] == v  # update/overwrite current key with new key value pair
#             else:  # if there is something already there and we do not want to overwrite it...
#                 continue  # continue to next key-value pair
#         else:  # if there's no key matching in this section, add key-value pair without further inspection
#             config_dict[config_section][k] == v
#
#     # write out this updated configuration dict to replace the previous configuration file
#     config_filename = flag_multiple_files(output_path, search_for=f'*{PIPELINE_CONFIG_HANDLE}*{CONFIG_FILETYPE}')
#
#     with open(config_filename, 'wt') as fout:
#         fout.write(tomlkit.dumps(config_dict))
#
#     return None

# locate relevant mapping file for demultiplexing
# moved down because it requires the import_config_as_dict function
def find_mapping_file(path_to_mapping, path_to_plex, path_to_config, **kwargs):
    '''
    Locate the barcode mapping file(s) for demultiplexing.

    Attempts to automatically detect the location of the mapping file
    corresponding to the multiplexed files located in the sorted
    directory of sequences to demultiplex. It inspects the name of the files
    that were sorted into the needs_demux folder of the sequences folder, and
    identifies the name of the sequencing run. It then will go into the mapping
    file directory in the config folder, mapping-files, and looks for the mapping
    file that shares that same sequencing run name. If multiple mapping files
    are detected for a single sequencing run, the user will be prompted to select
    the file that should be used. If no matching mapping file is detected, it
    will throw an error and exit the pipeline.
    :param path_to_mapping: defaults to expected location within structured
    file directories; otherwise can specify the path as a Path object
    :param path_to_plex: defaults to expected location within structured
    file directories; otherwise can specify the path to the mapping file(s) as
    a Path object
    :param kwargs: specify the bioinfo-settings configuration file dictionary
    if one is already loaded into the environment; if none is provided, the
    function will open the configuration file to create a new instance of the
    dictionary; adding a dictionary, if available, will likely be quicker so it
    is recommended to do so if possible
    :return: if found, returns mapping file(s) as a list
    '''
    assert is_pathclass(path_to_mapping)
    assert is_pathclass(path_to_plex)

    # get the multiplexed file's queue ID, which the UO sequencing core uses as the seq run identifier in the filename
    multiplex_ids = [file.stem.split('.')[0] for file in path_to_plex.glob('*.fast*')]

    # look up the multiplex IDs in the bioinfo-settings configuration file for its corresponding CliMush run name
    kwargs_keys = list(kwargs.keys())
    if kwargs_keys > 0:
        if kwargs_keys > 1:
            msg = f'ERROR: Too many keyword arguments (**kwargs) used. Only one kwarg is accepted and should have the '\
                  f'name of the configuration dictionary as the value to a key with any name.\n'
            print(msg)
            return exit_process(message=msg)
        else:
            config_multiplex_dict = kwargs[kwargs_keys[0]]['pacbio-multiplex-ids']
    else:
        config_multiplex_dict = import_config_as_dict(file_handle=path_to_config)['pacbio-multiplex-ids']

    # get the CliMush run name(s) and create regex to search for any of these at start of mapping file name
    climush_ids = [config_multiplex_dict[q] for q in multiplex_ids]

    # find the corresponding barcode mapping files for the run name(s) in the barcode-mapping dir in config
    mapping_file_names = []
    for id in climush_ids:
        match = list(path_to_mapping.glob(f'*{id}*'))
        if len(match) == 1:
            mapping_file_names.append(match[0])
        elif len(match) == 0:
            msg = f'ERROR: No mapping files matching the CliMush run name {id} were located in {path_to_mapping}. ' \
                  f'Check the mapping-files directory in the config folder, then retry.\n'
            print(msg)
            return exit_process(message=msg)
        else:
            msg = f'ERROR: {len(match)} mapping files were located for CliMush run name {id}. Please select the number ' \
                  f'of the correctly corresponding barcode mapping file for this sample:\n'
            print(msg)
            mapping_file_names.append(prompt_print_options(match))

    return mapping_file_names

def sort_input_files(filepath_dict, to_sort='main'):
    '''
    Sort the input sequencing files in their respective sequence
    platform subdirectories, or into subdirectories for required
    pre-processing.
    :param input_file_path: Path object to the directory containing
    the input sequences.
    :return: will print a count of how many files were sorted and
    into which folders, and then prompt to next steps if in
    interactive, or automatically carry out next steps if in
    automated.
    '''

    # UPDATE TO CHECK FOR ALL POSSIBLE DELIMITERS?
    def follows_naming_convention(filepath):
        if re.search(ANY_PLATFORM_REGEX, filepath.stem, re.I):
            return True
        else:
            return False

    def is_demultiplexed(filepath):
        if follows_naming_convention(filepath) or re.search('^[A-Z]+', filepath.stem):
            return True
        else:
            return False

    def is_pacbio(filepath):
        if re.search('^pacbio|^\d{4}', filepath.stem):
            return True
        else:
            return False

    def is_illumina(filepath):
        if re.search('^illumina|^[A-Z]+', filepath.stem):
            return True
        else:
            return False

    def is_sanger(filepath):
        if re.search('^sanger', filepath.stem):
            return True
        else:
            return False

    # I guess converting a generator to a list (like in print statement below) exhausts it? so then loop won't work
    files_to_sort = list(filepath_dict['sequences'][to_sort].glob('*.fast*'))

    print(f'Sorting {len(files_to_sort)} sequences files...\n')

    for file in files_to_sort:
        if is_pacbio(file):
            if is_demultiplexed(file):
                if follows_naming_convention(file):
                    move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['pacbio']))
                else:
                    move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['rename']))
            else:
                move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['demux']))
        elif is_illumina(file):
            if is_demultiplexed(file):
                if follows_naming_convention(file):
                    move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['illumina']))
                else:
                    move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['rename']))
            else:
                move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['demux']))
        elif is_sanger(file):
            if is_demultiplexed(file):
                if follows_naming_convention(file):
                    move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['sanger']))
                else:
                    move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['rename']))
            else:
                move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['demux']))
        else:
            move_file(source_path=file, dest_path=mkdir_exist_ok(new_dir=filepath_dict['sequences']['unclear']))

    return None

